from typing import TYPE_CHECKING

import logging
import random

import networkx
import numpy as np
from shapely.geometry import Point

from src.model.basic_functions import generate_weights, \
    compute_distance_from_arcmin, rescale_values

from src.agents.agent import Agent, Agents
from src.network.commercial_link import CommercialLink
from src.network.mrio import import_label

if TYPE_CHECKING:
    from src.agents.country import Countries
    from src.network.sc_network import ScNetwork
    from src.network.transport_network import TransportNetwork

EPSILON = 1e-6


class Firm(Agent):

    def __init__(self, pid, od_point=0, sector=0, sector_type=None, main_sector=None, name=None, input_mix=None,
                 target_margin=0.2, utilization_rate=0.8,
                 importance=1, long=None, lat=None, geometry=None,
                 suppliers=None, clients=None, production=0, inventory_duration_target=1, inventory_restoration_time=1,
                 usd_per_ton=2864, capital_to_value_added_ratio=4):
        super().__init__(
            agent_type="firm",
            pid=pid,
            name=name,
            od_point=od_point,
            long=long,
            lat=lat
        )
        # Parameters depending on data
        self.usd_per_ton = usd_per_ton
        self.geometry = geometry
        self.importance = importance
        self.sector = sector
        self.sector_type = sector_type
        self.main_sector = main_sector
        self.input_mix = input_mix or {}

        # Free parameters
        if input_mix is None:
            self.inventory_duration_target = inventory_duration_target
        else:
            self.inventory_duration_target = {key: inventory_duration_target for key in input_mix.keys()}
        self.inventory_restoration_time = inventory_restoration_time
        self.eq_production_capacity = production / utilization_rate
        self.utilization_rate = utilization_rate
        self.target_margin = target_margin
        self.capital_to_value_added_ratio = capital_to_value_added_ratio

        # Parameters depending on supplier-buyer network
        self.suppliers = suppliers or {}
        self.clients = clients or {}

        # Parameters sets at initialization
        self.eq_finance = {"sales": 0, 'costs': {"input": 0, "transport": 0, "other": 0}}
        self.eq_profit = 0
        self.eq_price = 1
        self.eq_total_order = 0

        # Variables, all initialized
        self.production = production
        self.production_target = production
        self.production_capacity = production / utilization_rate
        self.current_production_capacity = production
        self.capital_initial = 0
        self.purchase_plan = {}
        self.purchase_plan_per_input = {}
        self.order_book = {}
        self.total_order = 0
        self.input_needs = {}
        self.rationing = 1
        self.eq_needs = {}
        self.current_inventory_duration = {}
        self.inventory = {}
        self.product_stock = 0
        self.profit = 0
        self.finance = {"sales": 0, 'costs': {"input": 0, "transport": 0, "other": 0}}
        self.delta_price_input = 0
        self.generalized_transport_cost = 0
        self.usd_transported = 0
        self.tons_transported = 0
        self.tonkm_transported = 0

        # Disruption
        self.capital_destroyed = 0
        self.remaining_disrupted_time = 0
        self.production_capacity_reduction = 0
        self.capital_demanded = 0
        self.reconstruction_demand = 0
        self.reconstruction_produced = 0

    def reset_variables(self):
        self.eq_finance = {"sales": 0, 'costs': {"input": 0, "transport": 0, "other": 0}}
        self.eq_profit = 0
        self.eq_price = 1
        self.production = 0
        self.production_target = 0
        self.production_capacity = self.eq_production_capacity
        self.purchase_plan = {}
        self.order_book = {}
        self.total_order = 0
        self.input_needs = {}
        self.rationing = 1
        self.eq_needs = {}
        self.current_inventory_duration = {}
        self.inventory = {}
        self.product_stock = 0
        self.profit = 0
        self.finance = {"sales": 0, 'costs': {"input": 0, "transport": 0, "other": 0}}
        self.delta_price_input = 0
        self.generalized_transport_cost = 0
        self.usd_transported = 0
        self.tons_transported = 0
        self.tonkm_transported = 0
        self.capital_destroyed = 0

    def reset_indicators(self):
        pass

    def id_str(self):
        return super().id_str() + f" sector {self.sector}"

    def update_disrupted_production_capacity(self):
        is_back_to_normal = self.remaining_disrupted_time == 1  # Identify those who will be back to normal
        if self.remaining_disrupted_time > 0:  # Update the remaining time in disruption
            self.remaining_disrupted_time -= 1
        if is_back_to_normal:  # Update the remaining time in disruption
            self.production_capacity_reduction = 0
            logging.info(f'The production capacity of firm {self.pid} is back to normal')

    def disrupt_production_capacity(self, disruption_duration: int, reduction: float):
        self.remaining_disrupted_time = disruption_duration
        self.production_capacity_reduction = reduction
        logging.info(f'The production capacity of firm {self.pid} is reduced by {reduction} '
                     f'for {disruption_duration} time steps')

    def initialize_operational_variables(self, eq_production):
        self.production_target = eq_production
        self.production = self.production_target
        self.eq_production_capacity = self.production_target / self.utilization_rate
        self.production_capacity = self.eq_production_capacity
        self.evaluate_input_needs()
        self.eq_needs = self.input_needs
        self.inventory = {
            input_id: need * (1 + self.inventory_duration_target[input_id])
            for input_id, need in self.input_needs.items()
        }
        self.decide_purchase_plan(adaptive_inventories=False, adapt_weight_based_on_satisfaction=False)
        self.capital_initial = self.capital_to_value_added_ratio * eq_production

    def initialize_financial_variables(self, eq_production, eq_input_cost,
                                       eq_transport_cost, eq_other_cost):
        self.eq_finance['sales'] = eq_production
        self.eq_finance['costs']['input'] = eq_input_cost
        self.eq_finance['costs']['transport'] = eq_transport_cost
        self.eq_finance['costs']['other'] = eq_other_cost
        self.eq_profit = self.eq_finance['sales'] - sum(self.eq_finance['costs'].values())
        self.finance['sales'] = self.eq_finance['sales']
        self.finance['costs']['input'] = self.eq_finance['costs']['input']
        self.finance['costs']['transport'] = self.eq_finance['costs']['transport']
        self.finance['costs']['other'] = self.eq_finance['costs']['other']
        self.profit = self.eq_profit
        self.delta_price_input = 0

    def add_noise_to_geometry(self, noise_level=1e-5):
        self.geometry = Point(self.long + noise_level * random.uniform(0, 1),
                              self.lat + noise_level * random.uniform(0, 1))

    def distance_to_other(self, other_firm):
        if (self.od_point == -1) or (other_firm.od_point == -1):  # if virtual firms
            return 1
        else:
            return compute_distance_from_arcmin(self.long, self.lat, other_firm.long, other_firm.lat)

    def select_suppliers_from_data(self, graph, firm_list, inputed_supplier_links, output):

        for inputed_supplier_link in list(inputed_supplier_links.transpose().to_dict().values()):
            # Create an edge in the graph
            supplier_id = inputed_supplier_link['supplier_id']
            product_sector = inputed_supplier_link['product_sector']
            supplier_object = firm_list[supplier_id]
            graph.add_edge(supplier_object, self,
                           object=CommercialLink(
                               pid=str(supplier_id) + "->" + str(self.pid),
                               product=product_sector,
                               product_type=supplier_object.sector_type,
                               essential=inputed_supplier_link['is_essential'],
                               category='domestic_B2B',
                               supplier_id=supplier_id,
                               buyer_id=self.pid)
                           )
            # Associate a weight to the edge
            weight_in_input_mix = inputed_supplier_link['transaction'] / output
            graph[supplier_object][self]['weight'] = weight_in_input_mix
            # The firm saves the name of the supplier, its sector,
            # its weight among firm of the same sector (without I/O technical coefficient)
            total_input_same_sector = inputed_supplier_links.loc[
                inputed_supplier_links['product_sector'] == product_sector, "transaction"].sum()
            weight_among_same_product = inputed_supplier_link['transaction'] / total_input_same_sector
            self.suppliers[supplier_id] = {'sector': product_sector, 'weight': weight_among_same_product,
                                           "satisfaction": 1}

    def identify_suppliers(self, sector: str, firms, countries,
                           nb_suppliers_per_input: float, weight_localization: float,
                           firm_data_type: str, import_code: str):
        if firm_data_type == "mrio":
            if import_label in sector:  # case of countries
                supplier_type = "country"
                selected_supplier_ids = [sector[:3]]  # for countries, the id is extracted from the name
                supplier_weights = [1]

            else:  # case of firms
                supplier_type = "firm"
                potential_supplier_pids = [pid for pid, firm in firms.items() if firm.sector == sector]
                if sector == self.sector:
                    potential_supplier_pids.remove(self.pid)  # remove oneself
                if len(potential_supplier_pids) == 0:
                    raise ValueError(f"Firm {self.pid}: there should be one supplier for {sector}")
                # Choose based on importance
                prob_to_be_selected = np.array(rescale_values([firms[firm_pid].importance for firm_pid in
                                                               potential_supplier_pids]))
                prob_to_be_selected /= prob_to_be_selected.sum()
                selected_supplier_ids = np.random.choice(potential_supplier_pids,
                                                         p=prob_to_be_selected, size=1,
                                                         replace=False).tolist()
                supplier_weights = [1]

        else:
            if sector == import_code:
                supplier_type = "country"
                # Identify countries as suppliers if the corresponding sector does export
                importance_threshold = 1e-6
                potential_supplier_pid = [
                    pid
                    for pid, country in countries.items()
                    if country.supply_importance > importance_threshold
                ]
                importance_of_each = [
                    country.supply_importance
                    for country in countries.values()
                    if country.supply_importance > importance_threshold
                ]
                prob_to_be_selected = np.array(importance_of_each)
                prob_to_be_selected /= prob_to_be_selected.sum()

            # For the other types of inputs, identify the domestic suppliers, and
            # calculate their probability to be chosen, based on distance and importance
            else:
                supplier_type = "firm"
                potential_supplier_pid = [pid for pid, firm in firms.items() if firm.sector == sector]
                if sector == self.sector:
                    potential_supplier_pid.remove(self.pid)  # remove oneself
                if len(potential_supplier_pid) == 0:
                    raise ValueError(f"Firm {self.pid}: no potential supplier for input {sector}")
                distance_to_each = rescale_values([
                    self.distance_to_other(firms[firm_pid])
                    for firm_pid in potential_supplier_pid
                ])  # Compute distance to each of them (vol d oiseau)
                importance_of_each = rescale_values([firms[firm_pid].importance for firm_pid in
                                                     potential_supplier_pid])  # Get importance for each of them
                prob_to_be_selected = np.array(importance_of_each) / (np.array(distance_to_each) ** weight_localization)
                prob_to_be_selected /= prob_to_be_selected.sum()

            # Determine the number of supplier(s) to select. 1 or 2.
            if random.uniform(0, 1) < nb_suppliers_per_input - 1:
                nb_suppliers_to_choose = 2
                if nb_suppliers_to_choose > len(potential_supplier_pid):
                    nb_suppliers_to_choose = 1
            else:
                nb_suppliers_to_choose = 1

            # Select the supplier(s). It there is 2 suppliers, then we generate
            # random weight. It determines how much is bought from each supplier.
            selected_supplier_ids = np.random.choice(potential_supplier_pid,
                                                     p=prob_to_be_selected, size=nb_suppliers_to_choose,
                                                     replace=False).tolist()
            index_map = {supplier_id: position for position, supplier_id in enumerate(potential_supplier_pid)}
            selected_positions = [index_map[supplier_id] for supplier_id in selected_supplier_ids]
            selected_prob = [prob_to_be_selected[position] for position in selected_positions]
            supplier_weights = generate_weights(nb_suppliers_to_choose, selected_prob)

        return supplier_type, selected_supplier_ids, supplier_weights

    def select_suppliers(self, graph: "ScNetwork", firms: "Firms", countries: "Countries",
                         nb_suppliers_per_input: float, weight_localization: float,
                         firm_data_type: str, import_code: str):
        """
        The firm selects its suppliers.

        The firm checks its input mix to identify which type of inputs are needed.
        For each type of input, it selects the appropriate number of suppliers.
        Choice of suppliers is random, based on distance to eligible suppliers and
        their importance.

        If imports are needed, the firms select a country as supplier. Choice is
        random, based on the country's importance.

        Parameters
        ----------
        firm_data_type
        graph : networkx.DiGraph
            Supply chain graph
        firms : list of Firms
            Generated by createFirms function
        countries : list of Countries
            Generated by createCountries function
        nb_suppliers_per_input : float between 1 and 2
            Nb of suppliers per type of inputs. If it is a decimal between 1 and 2,
            some firms will have 1 supplier, other 2 suppliers, such that the
            average matches the specified value.
        weight_localization : float
            Give weight to distance when choosing supplier. The larger, the closer
            the suppliers will be selected.
        import_code : string
            Code that identify imports in the input mix.

        Returns
        -------
        int
            0

        """
        for sector_id, sector_weight in self.input_mix.items():

            # If it is imports, identify international suppliers and calculate
            # their probability to be chosen, which is based on importance.
            supplier_type, selected_supplier_ids, supplier_weights = self.identify_suppliers(sector_id,
                                                                                             firms, countries,
                                                                                             nb_suppliers_per_input,
                                                                                             weight_localization,
                                                                                             firm_data_type,
                                                                                             import_code)

            # For each new supplier, create a new CommercialLink in the supply chain network.
            # print(f"{self.id_str()}: for input {sector_id} I selected {len(selected_supplier_ids)} suppliers")
            for supplier_id in selected_supplier_ids:
                # Retrieve the appropriate supplier object from the id
                # If it is a country we get it from the country list
                # If it is a firm we get it from the firm list
                if supplier_type == "country":
                    supplier_object = [country for pid, country in countries.items() if pid == supplier_id][0]
                    link_category = 'import'
                    product_type = "imports"
                else:
                    supplier_object = firms[supplier_id]
                    link_category = 'domestic_B2B'
                    product_type = firms[supplier_id].sector_type
                # Create an edge in the graph
                graph.add_edge(supplier_object, self,
                               object=CommercialLink(
                                   pid=str(supplier_id) + "->" + str(self.pid),
                                   product=sector_id,
                                   product_type=product_type,
                                   category=link_category,
                                   supplier_id=supplier_id,
                                   buyer_id=self.pid)
                               )
                # Associate a weight, which includes the I/O technical coefficient
                supplier_weight = supplier_weights.pop()
                graph[supplier_object][self]['weight'] = sector_weight * supplier_weight
                # The firm saves the name of the supplier, its sector, its weight (without I/O technical coefficient)
                self.suppliers[supplier_id] = {'sector': sector_id, 'weight': supplier_weight, "satisfaction": 1}
                # The supplier saves the name of the client, its sector, and distance to it.
                # The share of sales cannot be calculated now
                distance = self.distance_to_other(supplier_object)
                supplier_object.clients[self.pid] = {'sector': self.sector, 'share': 0, 'transport_share': 0,
                                                     'distance': distance}

    def calculate_client_share_in_sales(self):
        # Only works if the order book was computed
        self.total_order = sum([order for client_pid, order in self.order_book.items()])
        total_qty_km = sum([
            info['distance'] * self.order_book[client_pid]
            for client_pid, info in self.clients.items()
        ])
        # self.total_B2B_order = sum([order for client_pid, order in self.order_book.items() if client_pid != -1])
        # If noone ordered to me, share is 0 (avoid division per 0)
        if self.total_order == 0:
            for client_pid, info in self.clients.items():
                info['share'] = 0
                info['transport_share'] = 0

        # If some clients ordered to me, but distance is 0 (no transport), then equal share of transport
        elif total_qty_km == 0:
            nb_active_clients = sum([order > 0 for client_pid, order in self.order_book.items()
                                     if client_pid != "reconstruction"])
            for client_pid, info in self.clients.items():
                info['share'] = self.order_book[client_pid] / self.total_order
                info['transport_share'] = 1 / nb_active_clients

        # Otherwise, standard case
        else:
            for client_pid, info in self.clients.items():
                info['share'] = self.order_book[client_pid] / self.total_order
                info['transport_share'] = self.order_book[client_pid] * self.clients[client_pid][
                    'distance'] / total_qty_km

    def aggregate_orders(self, log_info=False):
        self.total_order = sum([order for client_pid, order in self.order_book.items()])
        if log_info:
            if self.total_order == 0:
                logging.debug(f'Firm {self.pid} ({self.sector}): noone ordered to me')

    def decide_production_plan(self):
        self.production_target = max(0.0, self.total_order - self.product_stock)

    def calculate_price(self, graph):
        """
        Evaluate the relative increase in price due to changes in input price
        In addition, upon delivery, price will be adjusted for each client to reflect potential rerouting
        """
        if self.check_if_supplier_changed_price(graph):
            # One way to compute it is commented.
            #     self.delta_price_input = self.calculate_input_induced_price_change(graph)
            #     logging.debug('Firm '+str(self.pid)+': Input prices have changed, I set '+
            #     "my price to "+'{:.4f}'.format(self.eq_price*(1+self.delta_price_input))+
            #     " instead of "+str(self.eq_price))

            # I compute how much would be my input cost to produce one unit of output
            # if I had to buy the input at this price
            eq_unitary_input_cost, est_unitary_input_cost_at_current_price = self.get_input_costs(graph)
            # I scale this added cost to my total orders
            self.delta_price_input = est_unitary_input_cost_at_current_price - eq_unitary_input_cost
            if self.delta_price_input is np.nan:
                print(self.delta_price_input)
                print(est_unitary_input_cost_at_current_price)
                print(eq_unitary_input_cost)
            # added_input_cost = (est_unitary_input_cost_at_current_price - eq_unitary_input_cost) * self.total_order
            # self.delta_price_input = added_input_cost / self.total_order
            logging.debug('Firm ' + str(self.pid) + ': Input prices have changed, I set my price to ' +
                          '{:.4f}'.format(self.eq_price * (1 + self.delta_price_input)) +
                          " instead of " + str(self.eq_price))
        else:
            self.delta_price_input = 0

    def get_input_costs(self, graph):
        eq_unitary_input_cost = 0
        est_unitary_input_cost_at_current_price = 0
        for edge in graph.in_edges(self):
            eq_unitary_input_cost += graph[edge[0]][self]['object'].eq_price * graph[edge[0]][self]['weight']
            est_unitary_input_cost_at_current_price += graph[edge[0]][self]['object'].price * graph[edge[0]][self][
                'weight']
        return eq_unitary_input_cost, est_unitary_input_cost_at_current_price

    def evaluate_input_needs(self):
        self.input_needs = {
            input_pid: self.input_mix[input_pid] * self.production_target
            for input_pid, mix in self.input_mix.items()
        }

    def decide_purchase_plan(self, adaptive_inventories: bool, adapt_weight_based_on_satisfaction: bool):
        """
        If adaptive_inventories, it aims to come back to equilibrium inventories
        Else, it uses current orders to evaluate the target inventories
        """

        if adaptive_inventories:
            ref_input_needs = self.input_needs
        else:
            ref_input_needs = self.eq_needs

        # Evaluate the current safety days
        self.current_inventory_duration = {
            input_id: (evaluate_inventory_duration(ref_input_needs[input_id],
                                                   stock) if input_id in ref_input_needs.keys() else 0)
            for input_id, stock in self.inventory.items()
        }

        # Alert if there is less than a day of an input
        for input_id, inventory_duration in self.current_inventory_duration.items():
            if inventory_duration is not None:
                if inventory_duration < 1 - EPSILON:
                    # if -1 in self.clients.keys():
                    #     sales_to_hh = self.clients[-1]['share'] * self.production_target
                    # else:
                    #     sales_to_hh = 0
                    logging.debug(f"{self.id_str()} - Less than 1 day of inventory for input type {input_id}: "
                                  f"{inventory_duration} vs. {self.inventory_duration_target[input_id]}")

        # Evaluate purchase plan for each sector
        self.purchase_plan_per_input = {
            input_id: purchase_planning_function(need, self.inventory[input_id],
                                                 self.inventory_duration_target[input_id],
                                                 self.inventory_restoration_time)
            # input_id: purchase_planning_function(need, self.inventory[input_id],
            # self.inventory_duration_old, self.reactivity_rate)
            for input_id, need in ref_input_needs.items()
        }

        # Deduce the purchase plan for each supplier
        if adapt_weight_based_on_satisfaction:
            self.purchase_plan = {}
            for sector, need in self.purchase_plan_per_input.items():
                suppliers_from_this_sector = [pid for pid, supplier_info in self.suppliers.items()
                                              if supplier_info['sector'] == sector]
                total_satisfaction_suppliers = sum([supplier_info['satisfaction']
                                                    for pid, supplier_info in self.suppliers.items()
                                                    if pid in suppliers_from_this_sector])
                # print(self.id_str(), sector, need, "total_satisfaction_suppliers", total_satisfaction_suppliers)
                for supplier_id in suppliers_from_this_sector:
                    supplier_info = self.suppliers[supplier_id]
                    if total_satisfaction_suppliers < EPSILON:
                        self.purchase_plan[supplier_id] = need * supplier_info['weight']
                    else:
                        relative_satisfaction = supplier_info['satisfaction'] / total_satisfaction_suppliers
                        self.purchase_plan[supplier_id] = need * supplier_info['weight'] * relative_satisfaction

        else:
            self.purchase_plan = {
                supplier_id: self.purchase_plan_per_input[info['sector']] * info['weight']
                for supplier_id, info in self.suppliers.items()
            }

    def send_purchase_orders(self, sc_network: "ScNetwork"):
        for edge in sc_network.in_edges(self):
            supplier_id = edge[0].pid
            input_sector = edge[0].sector
            if supplier_id in self.purchase_plan.keys():
                quantity_to_buy = self.purchase_plan[supplier_id]
                if quantity_to_buy == 0:
                    logging.debug(f"{self.id_str()} - I am not planning to buy anything from supplier {supplier_id} "
                                  f"of sector {input_sector}. ")
                    if self.purchase_plan_per_input[input_sector] == 0:
                        logging.debug(f"{self.id_str()} - I am not planning to buy this input at all. "
                                      f"My needs is {self.input_needs[input_sector]}, "
                                      f"my inventory is {self.inventory[input_sector]}, "
                                      f"my inventory target is {self.inventory_duration_target[input_sector]} and "
                                      f"this input counts {self.input_mix[input_sector]} in my input mix.")
                    else:
                        logging.debug(f"But I plan to buy {self.purchase_plan_per_input[input_sector]} of this input")
            else:
                logging.error(f"{self.id_str()} - Supplier {supplier_id} is not in my purchase plan")
                quantity_to_buy = 0
            sc_network[edge[0]][self]['object'].order = quantity_to_buy

    def retrieve_orders(self, sc_network: "ScNetwork"):
        for edge in sc_network.out_edges(self):
            quantity_ordered = sc_network[self][edge[1]]['object'].order
            self.order_book[edge[1].pid] = quantity_ordered

    def add_reconstruction_order_to_order_book(self):
        self.order_book["reconstruction"] = self.reconstruction_demand

    def evaluate_capacity(self):
        if self.capital_destroyed > EPSILON:
            self.production_capacity_reduction = min(self.capital_destroyed / self.capital_initial, 1)
            logging.debug(f"{self.id_str()} - due to capital destruction, "
                          f"my production capacity is reduced by {self.production_capacity_reduction}")
        else:
            self.production_capacity_reduction = 0
        self.current_production_capacity = self.production_capacity * (1 - self.production_capacity_reduction)

    def incur_capital_destruction(self, amount: float):
        if amount > self.capital_initial:
            logging.warning(f"{self.id_str()} - initial capital is lower than destroyed capital "
                            f"({self.capital_initial} vs. {amount})")
            self.capital_destroyed = self.capital_initial
        else:
            self.capital_destroyed = amount

    def get_spare_production_potential(self):
        if len(self.input_mix) == 0:  # If no need for inputs
            potential_production = self.current_production_capacity
        else:
            # max_production = production_function(self.inventory, self.input_mix, mode)  # Max prod given inventory
            max_production = production_function(self.inventory, self.input_mix)  # Max prod given inventory
            potential_production = min([max_production, self.current_production_capacity])
        return max(0, potential_production - (self.total_order - self.product_stock))

    def produce(self, mode="Leontief"):
        # Produce
        if len(self.input_mix) == 0:  # If no need for inputs
            self.production = min([self.production_target, self.current_production_capacity])
        else:
            # max_production = production_function(self.inventory, self.input_mix, mode)  # Max prod given inventory
            max_production = production_function(self.inventory, self.input_mix, mode)  # Max prod given inventory
            self.production = min([max_production, self.production_target, self.current_production_capacity])

        # Add to stock of finished goods
        self.product_stock += self.production

        # Remove input used from inventories
        if mode == "Leontief":
            input_used = {input_id: self.production * mix for input_id, mix in self.input_mix.items()}
            self.inventory = {input_id: quantity - input_used[input_id] for input_id, quantity in
                              self.inventory.items()}
        else:
            raise ValueError("Wrong mode chosen")

    def calculate_input_induced_price_change(self, graph):
        """The firm evaluates the input costs of producing one unit of output if it had to buy the inputs at current
        price It is a theoretical cost, because in simulations it may use inventory
        """
        eq_theoretical_input_cost, current_theoretical_input_cost = self.get_input_costs(graph)
        input_cost_share = eq_theoretical_input_cost / 1
        relative_change = (current_theoretical_input_cost - eq_theoretical_input_cost) / eq_theoretical_input_cost
        return relative_change * input_cost_share / (1 - self.target_margin)

    def check_if_supplier_changed_price(self, graph):  # firms could record the last price they paid their input
        for edge in graph.in_edges(self):
            if abs(graph[edge[0]][self]['object'].price - graph[edge[0]][self]['object'].eq_price) > 1e-6:
                return True
        return False

    def ration_quantity_to_deliver(self):
        # remove rationing as attribute
        pass

    def deliver_products(self, graph: "ScNetwork", transport_network: "TransportNetwork",
                         sectors_no_transport_network: list, rationing_mode: str, monetary_units_in_model: str,
                         cost_repercussion_mode: str, price_increase_threshold: float, capacity_constraint: bool,
                         transport_cost_noise_level: float):

        # Do nothing if no orders
        if self.total_order == 0:
            return 0

        # Otherwise compute rationing factor
        self.rationing = self.product_stock / self.total_order
        # Check the case in which the firm has too much product to sale
        # It should not happen, hence a warning
        if self.rationing > 1 + EPSILON:
            logging.warning(f'Firm {self.pid}: I have produced too much. {self.product_stock} vs. {self.total_order}')
            self.rationing = 1
            quantity_to_deliver = {buyer_id: order for buyer_id, order in self.order_book.items()}
        # If rationing factor is 1, then it delivers what was ordered
        elif abs(self.rationing - 1) < EPSILON:
            quantity_to_deliver = {buyer_id: order for buyer_id, order in self.order_book.items()}
        # If rationing occurs, then two rationing behavior: equal or household_first
        else:
            logging.debug(f'Firm {self.pid}: I have to ration my clients by {(1 - self.rationing) * 100:.2f}%')
            # If equal, simply apply rationing factor
            if rationing_mode == "equal":
                quantity_to_deliver = {buyer_id: order * self.rationing for buyer_id, order in self.order_book.items()}

            # elif rationing_mode == "household_first": TODO: redo, does not work anymore
            #     if -1 not in self.order_book.keys():
            #         quantity_to_deliver = {buyer_id: order * self.rationing for buyer_id, order in
            #                                self.order_book.items()}
            #     elif len(self.order_book.keys()) == 1:  # only households order to this firm
            #         quantity_to_deliver = {-1: self.total_order}
            #     else:
            #         order_households = self.order_book[-1]
            #         if order_households < self.product_stock:
            #             remaining_product_stock = self.product_stock - order_households
            #             if (self.total_order - order_households) <= 0:
            #                 logging.warning("Firm " + str(self.pid) + ': ' + str(self.total_order - order_households))
            #             rationing_for_business = remaining_product_stock / (self.total_order - order_households)
            #             quantity_to_deliver = {buyer_id: order * rationing_for_business for buyer_id, order in
            #                                    self.order_book.items() if buyer_id != -1}
            #             quantity_to_deliver[-1] = order_households
            #         else:
            #             quantity_to_deliver = {buyer_id: 0 for buyer_id, order in self.order_book.items() if
            #                                    buyer_id != -1}
            #             quantity_to_deliver[-1] = self.product_stock
            else:
                raise ValueError('Wrong rationing_mode chosen')
        # We initialize transport costs, it will be updated for each shipment
        self.finance['costs']['transport'] = 0
        self.generalized_transport_cost = 0
        self.usd_transported = 0
        self.tons_transported = 0
        self.tonkm_transported = 0

        # For each client, we define the quantity to deliver then send the shipment
        for edge in graph.out_edges(self):
            if graph[self][edge[1]]['object'].order == 0:
                logging.debug(f"{self.id_str()} - {graph[self][edge[1]]['object'].buyer_id} "
                              f"is my client but did not order")
                continue
            graph[self][edge[1]]['object'].delivery = quantity_to_deliver[edge[1].pid]
            graph[self][edge[1]]['object'].delivery_in_tons = \
                Firm.transformUSD_to_tons(quantity_to_deliver[edge[1].pid], monetary_units_in_model, self.usd_per_ton)

            # If the client is B2C (applied only we had one single representative agent for all households)
            if edge[1].pid == -1:
                self.deliver_without_infrastructure(graph[self][edge[1]]['object'])
            # If this is service flow, deliver without infrastructure
            elif self.sector_type in sectors_no_transport_network:
                self.deliver_without_infrastructure(graph[self][edge[1]]['object'])
            # otherwise use infrastructure
            else:
                self.send_shipment(graph[self][edge[1]]['object'], transport_network, monetary_units_in_model,
                                   cost_repercussion_mode, price_increase_threshold, capacity_constraint,
                                   transport_cost_noise_level)

        # For reconstruction orders, we register it
        # if self.sector == "CON":
        #     print(quantity_to_deliver)
        if "reconstruction" in quantity_to_deliver.keys():
            self.reconstruction_produced = quantity_to_deliver['reconstruction']

    def deliver_without_infrastructure(self, commercial_link):
        """ The firm deliver its products without using transportation infrastructure
        This case applies to service firm and to households
        Note that we still account for transport cost, proportionally to the share of the clients
        Price can be higher than 1, if there are changes in price inputs
        """
        commercial_link.price = commercial_link.eq_price * (1 + self.delta_price_input)
        self.product_stock -= commercial_link.delivery
        self.finance['costs']['transport'] += (self.clients[commercial_link.buyer_id]['share'] *
                                               self.eq_finance['costs']['transport'])

    def send_shipment(self, commercial_link: "CommercialLink", transport_network: "TransportNetwork",
                      monetary_units_in_model: str, cost_repercussion_mode: str, price_increase_threshold: float,
                      capacity_constraint: bool, transport_cost_noise_level: float):

        monetary_unit_factor = {
            "mUSD": 1e6,
            "kUSD": 1e3,
            "USD": 1
        }
        factor = monetary_unit_factor[monetary_units_in_model]
        # print("send_shipment", 0 in transport_network.nodes)
        """Only apply to B2B flows 
        """
        if len(commercial_link.route) == 0:
            raise ValueError(f"{self.id_str()} - commercial link {commercial_link.pid} "
                             f"(qty {commercial_link.order:.02} is not associated to any route, "
                             f"I cannot send any shipment to the client")

        if self.check_route_availability(commercial_link, transport_network, 'main') == 'available':
            # If the normal route is available, we can send the shipment as usual
            # and pay the usual price
            commercial_link.price = commercial_link.eq_price * (1 + self.delta_price_input)
            commercial_link.current_route = 'main'
            transport_network.transport_shipment(commercial_link, capacity_constraint)
            self.product_stock -= commercial_link.delivery
            self.generalized_transport_cost += (commercial_link.route_time_cost
                                                + commercial_link.delivery_in_tons * commercial_link.route_cost_per_ton)
            self.usd_transported += commercial_link.delivery
            self.tons_transported += commercial_link.delivery_in_tons
            self.tonkm_transported += commercial_link.delivery_in_tons * commercial_link.route_length
            self.finance['costs']['transport'] += \
                self.clients[commercial_link.buyer_id]['share'] \
                * self.eq_finance['costs']['transport']
            return 0

        # If there is an alternative route already discovered,
        # and if this alternative route is available, then we use it
        if (len(commercial_link.alternative_route) > 0) \
                & (self.check_route_availability(commercial_link, transport_network, 'alternative') == 'available'):
            route = commercial_link.alternative_route
        # Otherwise we need to discover a new one
        else:
            route = self.discover_new_route(commercial_link, transport_network,
                                            capacity_constraint, transport_cost_noise_level)

        # If the alternative route is available, or if we discovered one, we proceed
        if route is not None:
            commercial_link.current_route = 'alternative'
            # Calculate contribution to generalized transport cost, to usd/tons/(ton*kms) transported
            self.generalized_transport_cost += (commercial_link.alternative_route_time_cost
                                                + Firm.transformUSD_to_tons(commercial_link.delivery,
                                                                            monetary_units_in_model, self.usd_per_ton)
                                                * commercial_link.alternative_route_cost_per_ton)
            self.usd_transported += commercial_link.delivery
            self.tons_transported += commercial_link.delivery_in_tons
            self.tonkm_transported += commercial_link.delivery_in_tons * commercial_link.alternative_route_length

            # We translate this real cost into transport cost
            if cost_repercussion_mode == "type1":  # relative cost change with actual bill
                # Calculate relative increase in routing cost
                new_transport_bill = commercial_link.delivery_in_tons * commercial_link.alternative_route_cost_per_ton
                normal_transport_bill = commercial_link.delivery_in_tons * commercial_link.route_cost_per_ton
                # print(normal_transport_bill, new_transport_bill)
                relative_cost_change = max(new_transport_bill - normal_transport_bill, 0) / normal_transport_bill
                # If switched transport mode, add switching cost
                # switching_cost = 0.5
                # if commercial_link.alternative_route_mode != commercial_link.route_mode:
                #     relative_cost_change = relative_cost_change + switching_cost
                # Translate that into an increase in transport costs in the balance sheet
                self.finance['costs']['transport'] += \
                    self.eq_finance['costs']['transport'] \
                    * self.clients[commercial_link.buyer_id]['transport_share'] \
                    * (1 + relative_cost_change)
                relative_price_change_transport = \
                    self.eq_finance['costs']['transport'] \
                    * relative_cost_change \
                    / ((1 - self.target_margin) * self.eq_finance['sales'])
                # Calculate the relative price change, including any increase due to the prices of inputs
                total_relative_price_change = self.delta_price_input + relative_price_change_transport
                commercial_link.price = commercial_link.eq_price * (1 + total_relative_price_change)

            elif cost_repercussion_mode == "type2":  # actual repercussion de la bill
                added_cost_usd_per_ton = max(commercial_link.alternative_route_cost_per_ton -
                                             commercial_link.route_cost_per_ton,
                                             0)
                added_cost_usd_per_musd = added_cost_usd_per_ton / (self.usd_per_ton / factor)
                added_cost_usd_per_musd = added_cost_usd_per_musd / factor
                added_transport_bill = added_cost_usd_per_musd * commercial_link.delivery
                self.finance['costs']['transport'] += \
                    self.eq_finance['costs']['transport'] + added_transport_bill
                commercial_link.price = (commercial_link.eq_price
                                         + self.delta_price_input
                                         + added_cost_usd_per_musd)
                relative_price_change_transport = \
                    commercial_link.price / (commercial_link.eq_price + self.delta_price_input) - 1
                if (commercial_link.price is None) or (commercial_link.price is np.nan):
                    raise ValueError("Price should be a float, it is " + str(commercial_link.price))

                cost_increase = (commercial_link.alternative_route_cost_per_ton
                                 - commercial_link.route_cost_per_ton) / commercial_link.route_cost_per_ton

                logging.debug(f"Firm {self.pid}"
                              f": qty {commercial_link.delivery_in_tons} tons"
                              f" increase in route cost per ton {cost_increase}"
                              f" increased bill mUSD {added_cost_usd_per_musd * commercial_link.delivery}"
                              )

            elif cost_repercussion_mode == "type3":
                relative_cost_change = (commercial_link.alternative_route_time_cost
                                        - commercial_link.route_time_cost) / commercial_link.route_time_cost
                self.finance['costs']['transport'] += (self.eq_finance['costs']['transport']
                                                       * self.clients[commercial_link.buyer_id]['share']
                                                       * (1 + relative_cost_change))
                relative_price_change_transport = (
                        self.eq_finance['costs']['transport']
                        * relative_cost_change
                        / ((1 - self.target_margin) * self.eq_finance['sales']))

                total_relative_price_change = self.delta_price_input + relative_price_change_transport
                commercial_link.price = commercial_link.eq_price * (1 + total_relative_price_change)
            else:
                raise NotImplementedError(f"Type {cost_repercussion_mode} not implemented")

            # If the increase in transport is larger than 2, then we do not deliver the goods
            if relative_price_change_transport > price_increase_threshold:
                logging.info(f"{self.id_str()}: found an alternative route to {commercial_link.buyer_id} "
                             f"but it is costlier by {100 * relative_price_change_transport:.2f}%, "
                             f"price would be {commercial_link.price:.4f} "
                             f"instead of {commercial_link.eq_price * (1 + self.delta_price_input):.4f}"
                             f"so I decide not to send it now.")
                commercial_link.price = commercial_link.eq_price
                commercial_link.current_route = 'none'
                # commercial_link.delivery = 0
            # Otherwise, we deliver
            else:
                transport_network.transport_shipment(commercial_link, capacity_constraint)
                self.product_stock -= commercial_link.delivery
                # Print information
                logging.info("Firm " + str(self.pid) + ": found an alternative route to " +
                             str(commercial_link.buyer_id) + ", it is costlier by " +
                             '{:.2f}'.format(100 * relative_price_change_transport) + "%, price is " +
                             '{:.4f}'.format(commercial_link.price) + " instead of " +
                             '{:.4f}'.format(commercial_link.eq_price * (1 + self.delta_price_input))
                             )
        # If we do not find a route, then we do not deliver
        else:
            logging.info(f"{self.id_str()}: because of disruption, there is no route between me "
                         f"and agent {commercial_link.buyer_id}")
            # We do not write how the input price would have changed
            commercial_link.price = commercial_link.eq_price
            commercial_link.current_route = 'none'
            # We do not pay the transporter, so we don't increment the transport cost
            # We set delivery to 0
            commercial_link.delivery = 0

    def discover_new_route(self, commercial_link: "CommercialLink", transport_network: "TransportNetwork",
                           account_capacity: bool, transport_cost_noise_level: float):
        origin_node = self.od_point
        destination_node = commercial_link.route[-1][0]
        route = self.choose_route(
            transport_network=transport_network.get_undisrupted_network(),
            origin_node=origin_node,
            destination_node=destination_node,
            capacity_constraint=account_capacity,
            transport_cost_noise_level=transport_cost_noise_level
        )
        # If we find a new route, we save it as the alternative one
        if route is not None:
            commercial_link.store_route_information(
                route=route,
                main_or_alternative="alternative"
            )
        return route

    def receive_service_and_pay(self, commercial_link: "CommercialLink"):
        super().receive_service_and_pay(commercial_link)
        quantity_delivered = commercial_link.delivery
        self.inventory[commercial_link.product] += quantity_delivered

    def update_indicator(self, quantity_delivered: float, price: float, commercial_link: "CommercialLink"):
        super().update_indicator(quantity_delivered, price, commercial_link)
        commercial_link.calculate_fulfilment_rate()
        self.suppliers[commercial_link.supplier_id]['satisfaction'] = commercial_link.fulfilment_rate

    def evaluate_profit(self, graph):
        # Collect all payments received
        self.finance['sales'] = sum([
            graph[self][edge[1]]['object'].payment
            for edge in graph.out_edges(self)
        ])
        # Collect all payments made
        self.finance['costs']['input'] = sum([
            graph[edge[0]][self]['object'].payment
            for edge in graph.in_edges(self)
        ])
        # Compute profit
        self.profit = (self.finance['sales']
                       - self.finance['costs']['input']
                       - self.finance['costs']['other']
                       - self.finance['costs']['transport'])
        # Compute Margins
        expected_gross_margin_no_transport = 1 - sum(list(self.input_mix.values()))
        if self.finance['sales'] > EPSILON:
            realized_gross_margin_no_transport = ((self.finance['sales'] - self.finance['costs']['input'])
                                                  / self.finance['sales'])
            realized_margin = self.profit / self.finance['sales']
        else:
            realized_gross_margin_no_transport = 0
            realized_margin = 0

        # Log discrepancies
        if abs(realized_gross_margin_no_transport - expected_gross_margin_no_transport) > 1e-6:
            logging.debug('Firm ' + str(self.pid) + ': realized gross margin without transport is ' +
                          '{:.3f}'.format(realized_gross_margin_no_transport) + " instead of " +
                          '{:.3f}'.format(expected_gross_margin_no_transport))

        if abs(realized_margin - self.target_margin) > 1e-6:
            logging.debug('Firm ' + str(self.pid) + ': my margin differs from the target one: ' +
                          '{:.3f}'.format(realized_margin) + ' instead of ' + str(self.target_margin))


class Firms(Agents):
    def __init__(self, agent_list=None):
        super().__init__(agent_list)

    def filter_by_sector(self, sector):
        filtered_agents = Firms()
        for agent in self.values():
            if agent.sector == sector:
                filtered_agents[agent.pid] = agent
        return filtered_agents

    def retrieve_orders(self, sc_network: "ScNetwork"):
        for firm in self.values():
            firm.retrieve_orders(sc_network)

    def plan_production(self, sc_network: "ScNetwork", propagate_input_price_change: bool = True):
        for firm in self.values():
            firm.evaluate_capacity()
            firm.aggregate_orders(log_info=True)
            firm.decide_production_plan()
            if propagate_input_price_change:
                firm.calculate_price(sc_network)

    def plan_purchase(self, adaptive_inventories: bool, adapt_weight_based_on_satisfaction: bool):
        for firm in self.values():
            firm.evaluate_input_needs()
            firm.decide_purchase_plan(adaptive_inventories, adapt_weight_based_on_satisfaction)  # mode="reactive"

    def produce(self):
        for firm in self.values():
            firm.produce()

    def evaluate_profit(self, sc_network: "ScNetwork"):
        for firm in self.values():
            firm.evaluate_profit(sc_network)

    def update_disrupted_production_capacity(self):
        for firm in self.values():
            firm.update_disrupted_production_capacity()

    def get_disrupted(self, firm_id_duration_reduction_dict: dict):
        for firm in self.values():
            if firm.pid in list(firm_id_duration_reduction_dict.keys()):
                firm.disrupt_production_capacity(
                    firm_id_duration_reduction_dict[firm.pid]['duration'],
                    firm_id_duration_reduction_dict[firm.pid]['reduction']
                )


def production_function(inputs, input_mix, function_type="Leontief"):
    # Leontief
    if function_type == "Leontief":
        try:
            return min([inputs[input_id] / input_mix[input_id] for input_id, val in input_mix.items()])
        except KeyError:
            return 0

    else:
        raise ValueError("Wrong mode selected")


def purchase_planning_function(estimated_need: float, inventory: float, inventory_duration_target: float,
                               inventory_restoration_time: float):
    """Decide the quantity of each input to buy according to a dynamical rule
    """
    # target_inventory = (1 + inventory_duration_target) * estimated_need
    target_inventory = inventory_duration_target * estimated_need
    # if inventory >= target_inventory + estimated_need:
    #     return 0
    # elif inventory >= target_inventory:
    #     return target_inventory + estimated_need - inventory
    # else:
    #     # return (1 - 1 / inventory_restoration_time) * estimated_need + inventory_restoration_time * (
    #     #         estimated_need + target_inventory - inventory)
    return max(0.0, estimated_need + 1 / inventory_restoration_time * (target_inventory - inventory))


def evaluate_inventory_duration(estimated_need, inventory):
    if estimated_need == 0:
        return None
    else:
        return inventory / estimated_need - 1



import logging

import networkx as nx
import pandas as pd

from src.agents.firm import Firm
from src.model.basic_functions import add_or_append_to_dict


class ScNetwork(nx.DiGraph):

    def access_commercial_link(self, edge):
        return self[edge[0]][edge[1]]['object']

    def calculate_io_matrix(self):
        io = {}
        for supplier, buyer, data in self.edges(data=True):
            commercial_link = data['object']
            if commercial_link.category == "domestic_B2C":
                add_or_append_to_dict(io, (supplier.sector, 'final_demand'), commercial_link.order)
            elif commercial_link.category == "export":
                add_or_append_to_dict(io, (supplier.sector, 'export'), commercial_link.order)
            elif commercial_link.category == "domestic_B2B":
                add_or_append_to_dict(io, (supplier.sector, buyer.sector), commercial_link.order)
            elif commercial_link.category == "import_B2C":
                add_or_append_to_dict(io, ("IMP", 'final_demand'), commercial_link.order)
            elif commercial_link.category == "import":
                add_or_append_to_dict(io, ("IMP", buyer.sector), commercial_link.order)
            elif commercial_link.category == "transit":
                pass
            else:
                raise KeyError('Commercial link categories should be one of domestic_B2B, '
                               'domestic_B2C, export, import, import_B2C, transit')

        io_table = pd.Series(io).unstack().fillna(0)
        return io_table

    def generate_edge_list(self):
        edge_list = [(source.pid, source.agent_type, source.od_point, target.pid, target.agent_type, target.od_point)
                     for source, target in self.edges()]
        edge_list = pd.DataFrame(edge_list)
        edge_list.columns = ['source_id', 'source_type', 'source_od_point',
                             'target_id', 'target_type', 'target_od_point']
        return edge_list

    def identify_firms_without_clients(self):
        return [node for node in self.nodes() if (self.out_degree(node) == 0) and isinstance(node, Firm)]

    def remove_useless_commercial_links(self):
        firms_without_clients = self.identify_firms_without_clients()
        # print(firms_without_clients)
        logging.info(f"There are {len(firms_without_clients)} firms without clients. Removing associated links")
        for firm_without_clients in firms_without_clients:
            suppliers = [edge[0] for edge in self.in_edges(firm_without_clients)]
            for supplier in suppliers:
                self.remove_edge(supplier, firm_without_clients)
                del supplier.clients[firm_without_clients.pid]
                del firm_without_clients.suppliers[supplier.pid]
        # logging.info(f"There remain {len(self.identify_firms_without_clients())} firms without clients.")
        # print(self.identify_firms_without_clients())



import logging

import geopandas
import geopandas as gpd
import pandas as pd
import yaml
from shapely.geometry import Point

from src.network.transport_network import TransportNetwork


def load_transport_data(filepaths, transport_params, transport_mode, transport_cost_data, time_resolution,
                        additional_roads=None):
    # Determines whether there are nodes and/or edges to load
    any_edge = True
    any_node = True
    if transport_mode == "multimodal":
        any_node = False

    # Load nodes
    if any_node:
        nodes = gpd.read_file(filepaths[transport_mode + '_nodes'])
        # if 'index' in nodes.columns: #remove any column named "index". Seems to create issues
        #     nodes = nodes.drop('index', axis=1)
        nodes.index = nodes['id']
        nodes.index.name = 'index'
        nodes['type'] = transport_mode

    # Load edges
    if any_edge:
        edges = gpd.read_file(filepaths[transport_mode + '_edges'])
        # if 'index' in edges.columns: #remove any column named "index". Seems to create issues
        #     edges = edges.drop('index', axis=1)
        edges.index = edges['id']
        edges.index.name = 'index'
        edges['type'] = transport_mode

        # Add additional road edges, if any
        if (transport_mode == "roads") and additional_roads:
            offset_edge_id = edges['id'].max() + 1
            new_road_edges = gpd.read_file(filepaths['extra_roads_edges'])
            # if 'index' in new_road_edges.columns: #remove any column named "index". Seems to create issues
            #     new_road_edges = new_road_edges.drop('index', axis=1)
            new_road_edges['id'] = new_road_edges['id'] + offset_edge_id
            new_road_edges.index = new_road_edges['id']
            new_road_edges['type'] = 'road'
            edges = pd.concat([edges, new_road_edges], ignore_index=False,
                              verify_integrity=True, sort=False)

        # Compute how much it costs to transport one USD worth of good on each edge
        edges = compute_cost_travel_time_edges(edges, transport_params, transport_mode, transport_cost_data)

        # Adapt capacity (given in tons per day) to time resolution
        time_resolution_in_days = {'day': 1, 'week': 7, 'month': 365.25/12, 'year': 365.25}
        edges['capacity'] = pd.to_numeric(edges['capacity'], errors="coerce")
        edges['capacity'] = edges['capacity'] * time_resolution_in_days[time_resolution]

        # When there is no capacity, it means that there is no limitation
        unlimited_capacity = 1e9 * time_resolution_in_days[time_resolution]  # tons per year
        no_capacity_cond = (edges['capacity'].isnull()) | (edges['capacity'] == 0)
        edges.loc[no_capacity_cond, 'capacity'] = unlimited_capacity
        # dic_capacity = {
        #     "roads": 1000000*52,
        #     "railways": 20000*52,
        #     "waterways": 40000*52,
        #     "maritime": 1e12*52,
        #     "multimodal": 1e12*52
        # }
        # edges['capacity'] = dic_capacity[transport_mode] / periods[time_resolution]
        # if (transport_mode == "roads"):
        #     edges.loc[edges['surface']=="unpaved", "capacity"] = 100000*52 / periods[time_resolution]

    # Return nodes, edges, or both
    if any_node and any_edge:
        return nodes, edges
    if any_node and ~any_edge:
        return nodes
    if ~any_node and any_edge:
        return edges


def offset_ids(nodes, edges, offset_node_id, offset_edge_id):
    # offset node id
    nodes['id'] = nodes['id'] + offset_node_id
    nodes.index = nodes['id']
    nodes.index.name = "index"
    # offset edge id
    edges['id'] = edges['id'] + offset_edge_id
    edges.index = edges['id']
    edges.index.name = "index"
    # offset end1, end2
    edges['end1'] = edges['end1'] + offset_node_id
    edges['end2'] = edges['end2'] + offset_node_id

    return nodes, edges


def create_transport_network(transport_modes: list, filepaths: dict, transport_cost_data: dict, time_resolution: str,
                             extra_roads=False):
    """Create the transport network object

    It uses one shapefile for the nodes and another for the edges.
    Note that there are strong constraints on these files, in particular on their attributes.
    We can optionally use an additional edge shapefile, which contains extra road segments. Useful for scenario testing.

    Parameters
    ----------
    transport_cost_data
    extra_roads
    transport_modes : list
        List of transport modes to include, ['roads', 'railways', 'waterways', 'airways']

    filepaths : dic
        Dic of filepaths

    Returns
    -------
    TransportNetwork, transport_nodes geopandas.DataFrame, transport_edges geopandas.DataFrame
    """

    # Load transport parameters
    with open(filepaths['transport_parameters'], "r") as yaml_file:
        transport_params = yaml.load(yaml_file, Loader=yaml.FullLoader)

    # Create the transport network object
    logging.info('Creating transport network')
    if extra_roads:
        logging.info('Including extra roads')
    transport_network = TransportNetwork()
    # T.graph['unit_cost'] = transport_params['transport_cost_per_tonkm']

    # Load node and edge data
    # Load in the following order: roads, railways, waterways
    # Ids are adjusted to be unique
    # E.g., if roads and railways are included, the ids of railways nodes are offset such that
    # the first railway node id is the last road node id + 1
    # similarly for edges
    # the nodes ids in "end1", "end2" of edges are also offseted
    logging.debug('Loading roads data')
    nodes, edges = load_transport_data(filepaths, transport_params,
                                       transport_mode="roads",
                                       transport_cost_data=transport_cost_data,
                                       additional_roads=extra_roads,
                                       time_resolution=time_resolution)
    logging.info(f"Transport modes modeled: {transport_modes}")
    if "railways" in transport_modes:
        nodes, edges = add_transport_mode("railways", nodes, edges, filepaths, transport_params, transport_cost_data,
                                          time_resolution)

    if "waterways" in transport_modes:
        nodes, edges = add_transport_mode("waterways", nodes, edges, filepaths, transport_params, transport_cost_data,
                                          time_resolution)

    if "maritime" in transport_modes:
        nodes, edges = add_transport_mode("maritime", nodes, edges, filepaths, transport_params, transport_cost_data,
                                          time_resolution)

    if "airways" in transport_modes:
        nodes, edges = add_transport_mode("airways", nodes, edges, filepaths, transport_params, transport_cost_data,
                                          time_resolution)

    if len(transport_modes) >= 2:
        logging.debug('Loading multimodal data')
        multimodal_edges = load_transport_data(filepaths, transport_params, "multimodal", transport_cost_data,
                                               time_resolution=time_resolution)
        multimodal_edges = select_multimodal_edges_needed(multimodal_edges, transport_modes)
        logging.debug(str(multimodal_edges.shape[0]) + " multimodal edges")
        multimodal_edges = assign_endpoints(multimodal_edges, nodes)
        multimodal_edges['id'] = multimodal_edges['id'] + edges['id'].max() + 1
        multimodal_edges.index = multimodal_edges['id']
        multimodal_edges.index.name = "index"
        edges = pd.concat([edges, multimodal_edges], ignore_index=False,
                          verify_integrity=True, sort=False)
        logging.debug('Total nb of transport nodes: ' + str(nodes.shape[0]))
        logging.debug('Total nb of transport edges: ' + str(edges.shape[0]))

    # Check conformity
    if nodes['id'].duplicated().sum() > 0:
        raise ValueError('The following node ids are duplicated: ' +
                         nodes.loc[nodes['id'].duplicated(), "id"])
    if edges['id'].duplicated().sum() > 0:
        raise ValueError('The following edge ids are duplicated: ' +
                         edges.loc[edges['id'].duplicated(), "id"])
    edge_ends = set(edges['end1'].tolist() + edges['end2'].tolist())
    edge_ends_not_in_node_data = edge_ends - set(nodes['id'])
    if len(edge_ends_not_in_node_data) > 0:
        raise KeyError("The following node ids are given as 'end1' or 'end2' in edge data " + \
                       "but are not in the node data: " + str(edge_ends_not_in_node_data))

    # Load the nodes and edges on the transport network object
    logging.debug('Creating transport nodes and edges as a network')
    for road in edges['id']:
        transport_network.add_transport_edge_with_nodes(road, edges, nodes)

    # Only keep nodes that are used in edges (there may be some unconnected nodes in the data)
    nodes = nodes.loc[nodes['id'].isin(list(transport_network.nodes))]

    return transport_network, nodes, edges


def add_transport_mode(
        mode: str,
        nodes: geopandas.GeoDataFrame,
        edges: geopandas.GeoDataFrame,
        filepaths: dict,
        transport_params: dict,
        transport_cost_data: dict,
        time_resolution: str
):
    logging.debug(f'Loading {mode} data')
    new_mode_nodes, new_mode_edges = load_transport_data(filepaths, transport_params, mode, transport_cost_data,
                                                         time_resolution=time_resolution)
    logging.debug(f"{new_mode_nodes.shape[0]} {mode} nodes and {new_mode_edges.shape[0]} {mode} edges")
    new_mode_nodes, new_mode_edges = offset_ids(new_mode_nodes, new_mode_edges,
                                                offset_node_id=nodes['id'].max() + 1,
                                                offset_edge_id=edges['id'].max() + 1)
    nodes = pd.concat([nodes, new_mode_nodes], ignore_index=False,
                      verify_integrity=True, sort=False)
    edges = pd.concat([edges, new_mode_edges], ignore_index=False,
                      verify_integrity=True, sort=False)
    return nodes, edges


def select_multimodal_edges_needed(multimodal_edges, transport_modes):
    # Select multimodal edges between the specified transport modes
    modes = multimodal_edges['multimodes'].str.split('-', expand=True)
    boolean = modes.iloc[:, 0].isin(transport_modes) & modes.iloc[:, 1].isin(transport_modes)
    return multimodal_edges[boolean]


def assign_endpoints_one_edge(row, df_nodes):
    p1, p2 = get_endpoints_from_line(row['geometry'])
    id_closest_point1 = get_index_closest_point(p1, df_nodes)
    id_closest_point2 = get_index_closest_point(p2, df_nodes)
    row['end1'] = id_closest_point1
    row['end2'] = id_closest_point2
    return row


def assign_endpoints(df_links, df_nodes):
    return df_links.apply(lambda row: assign_endpoints_one_edge(row, df_nodes), axis=1)


def get_endpoints_from_line(linestring_obj):
    end1_coord = linestring_obj.coords[0]
    end2_coord = linestring_obj.coords[-1]
    return Point(*end1_coord), Point(*end2_coord)


def get_index_closest_point(point, df_with_points):
    distance_list = [point.distance(item) for item in df_with_points['geometry'].tolist()]
    return int(df_with_points.index[distance_list.index(min(distance_list))])


def compute_cost_travel_time_edges(edges: geopandas.GeoDataFrame, transport_params: dict,
                                   edge_type: str, transport_cost_data: dict) -> geopandas.GeoDataFrame:
    # A. Compute price per ton to be paid to transporter
    # A1. Cost per km
    if edge_type == "roads":
        # Differentiate between paved and unpaved roads
        if transport_cost_data['roads'] == "edge-specific":
            edges['cost_per_ton'] = edges['km'] * edges['transport_cost_per_tonkm']
        elif transport_cost_data['roads'] == "mode-specific":
            edges['cost_per_ton'] = edges['km'] * transport_params['transport_cost_per_tonkm']['roads']
        elif transport_cost_data['roads'] == "surface-specific":
            edges['cost_per_ton'] = edges['km'] * (
                    (edges['surface'] == 'unpaved') * transport_params['transport_cost_per_tonkm']['roads']['unpaved']
                    + (edges['surface'] == 'paved') * transport_params['transport_cost_per_tonkm']['roads']['paved']
            )
        else:
            raise ValueError("Parameters transport_cost_data for roads should be one of: "
                             "edge-specific, mode-specific, surface-specific")
    elif edge_type in ['railways', 'waterways', 'maritime', 'airways']:
        if transport_cost_data[edge_type] == "mode-specific":
            edges['cost_per_ton'] = edges['km'] * transport_params['transport_cost_per_tonkm'][edge_type]
        else:
            raise ValueError("Parameters transport_cost_data for railways, waterways, maritime, airways "
                             "can only be mode-specific")

    elif edge_type == "multimodal":
        edges['cost_per_ton'] = edges['multimodes'].map(transport_params['loading_cost_per_ton'])

    else:
        raise ValueError("'edge_type' should be 'roads', 'railways', 'waterways', 'airways', or 'multimodal'")

    # A2. Forwarding charges at borders
    if not edges['special'].isnull().all():
        boolean_custom_edge = edges['special'].str.contains("custom", na=False)
        edges.loc[boolean_custom_edge, "cost_per_ton"] = edges.loc[boolean_custom_edge, "type"] \
            .map(transport_params['custom_cost'])

    # B. Compute generalized cost of transport
    # B.1a. Compute travel time
    if edge_type == "roads":
        # Differentiate between paved and unpaved roads
        edges['travel_time'] = edges['km'] * (
                (edges['surface'] == 'unpaved') / transport_params['speeds']['roads']['unpaved'] +
                (edges['surface'] == 'paved') / transport_params['speeds']['roads']['paved']
        )
    elif edge_type in ['railways', 'waterways', 'maritime', 'airways']:
        edges['travel_time'] = edges['km'] / transport_params['speeds'][edge_type]

    elif edge_type == "multimodal":
        edges['travel_time'] = edges['km'] / edges['multimodes'].map(transport_params['loading_time'])

    else:
        raise ValueError("'edge_type' should be 'roads', 'railways', 'waterways', 'airways', or 'multimodal'")

    # B.1b. Add crossing border time
    if not edges['special'].isnull().all():
        boolean_custom_edge = edges['special'].str.contains("custom", na=False)
        edges.loc[boolean_custom_edge, "travel_time"] = edges.loc[boolean_custom_edge, "type"] \
                .map(transport_params['custom_time'])

    # B.2. Compute cost of travel time
    edges['cost_travel_time'] = edges['travel_time'] * transport_params["travel_cost_of_time"]

    # B.3. Compute variability cost
    if edge_type == "roads":
        # Differentiate between paved and unpaved roads
        edges['cost_variability'] = edges['cost_travel_time'] * \
            transport_params['variability_coef'] * (
                (edges['surface'] == 'unpaved') * transport_params["variability"]['roads']['unpaved']
                + (edges['surface'] == 'paved') * transport_params['variability']['roads']['paved']
            )

    elif edge_type in ['railways', 'waterways', 'maritime', 'airways']:
        edges['cost_variability'] = edges['cost_travel_time'] * \
                                    transport_params['variability_coef'] * \
                                    transport_params["variability"][edge_type]

    elif edge_type == "multimodal":
        edges['cost_variability'] = edges['cost_travel_time'] * \
                                    transport_params['variability_coef'] * \
                                    edges['multimodes'].map(transport_params['variability']['multimodal'])

    else:
        raise ValueError("'edge_type' should be 'roads', 'railways', 'waterways', 'airways', or 'multimodal'")

    # B.4. Finally, add up both cost to get the cost of time of each road edges
    edges['time_cost'] = edges['cost_travel_time'] + edges['cost_variability']

    # C. Compute an aggregate cost
    # It is "cost_per_ton" + "time_cost"
    edges['agg_cost'] = edges['time_cost'] / 2 + edges['cost_per_ton']
    return edges



# Import modules
import logging
import sys
import time
import paths
from src.model.caching_functions import generate_cache_parameters_from_command_line_argument
from src.parameters import Parameters
from src.simulation.handling_functions import check_script_call
from model.model import Model

# Start run
t0 = time.time()

# Check that the script is called correctly
check_script_call(sys.argv)

# Retrieve scope
scope = sys.argv[1]
logging.info(f'Simulation starting for {scope}')

# Generate cache parameters
cache_parameters = generate_cache_parameters_from_command_line_argument(sys.argv)

# Import parameters
parameters = Parameters.load_parameters(paths.PARAMETER_FOLDER, scope)

# Create the output folder and adjust logging behavior
if parameters.export_files:
    parameters.create_export_folder()
    parameters.export()
    logging.info(f'Output folder is {parameters.export_folder}')

parameters.adjust_logging_behavior()

# Initialize model
model = Model(parameters)
model.setup_transport_network(cached=cache_parameters['transport_network'])
if parameters.export_files:
    model.export_transport_nodes_edges()
model.setup_agents(cached=cache_parameters['agents'])
if parameters.export_files:
    model.export_agent_tables()
model.setup_sc_network(cached=cache_parameters['sc_network'])
model.set_initial_conditions()
model.setup_logistic_routes(cached=cache_parameters['logistic_routes'])

# Run model
if parameters.simulation_type == "initial_state":
    simulation = model.run_static()

elif parameters.simulation_type == "disruption":
    simulation = model.run_disruption()

else:
    raise ValueError('Unimplemented simulation type chosen')

if parameters.export_files:
    simulation.export_agent_data(parameters.export_folder)
    simulation.export_transport_network_data(model.transport_edges, parameters.export_folder)
    simulation.calculate_and_export_summary_result(model.sc_network, model.household_table,
                                                   parameters.monetary_units_in_model, parameters.export_folder)

logging.info("End of simulation")



import pandas as pd

from src.network.route import Route
from src.parameters import EPSILON


class CommercialLink(object):

    def __init__(self, pid=None, supplier_id=None, buyer_id=None, product=None,
                 product_type=None, category=None, order=0, delivery=0, payment=0, essential=True,
                 route=None):
        # Parameter
        self.pid = pid
        self.product = product  # sector of producing firm
        self.product_type = product_type  # service, manufacturing, etc. (=sector_type)
        self.category = category  # import, export, domestic_B2B, transit
        self.route = route or []  # node_id path of the transport network, as
        # [(node1, ), (node1, node2), (node2, ), (node2, node3), (node3, )]
        self.route_length = 1
        self.route_time_cost = 0
        self.route_cost_per_ton = 0
        self.supplier_id = supplier_id
        self.buyer_id = buyer_id
        self.eq_price = 1
        self.possible_transport_modes = "any"
        self.essential = essential

        # Variable
        self.current_route = 'main'
        self.order = order  # flows upstream
        self.delivery = delivery  # flows downstream. What is supposed to be delivered (if no transport pb)
        self.delivery_in_tons = delivery  # flows downstream. What is supposed to be delivered (if no transport pb)
        self.realized_delivery = delivery
        self.payment = payment  # flows upstream
        self.alternative_route = []
        self.alternative_route_length = 1
        self.alternative_route_time_cost = 0
        self.alternative_route_cost_per_ton = 0
        self.price = 1
        self.fulfilment_rate = 1  # ratio deliver / order

    def print_info(self):
        # print("\nCommercial Link from "+str(self.supplier_id)+" to "+str(self.buyer_id)+":")
        # print("route:", self.route)
        # print("alternative route:", self.alternative_route)
        # print("product:", self.product)
        # print("order:", self.order)
        # print("delivery:", self.delivery)
        # print("payment:", self.payment)
        attribute_to_print = [a for a in dir(self) if not a.startswith('__') and not callable(getattr(self, a))]
        for attribute in attribute_to_print:
            print(attribute + ": " + str(getattr(self, attribute)))

    def reset_variables(self):
        # Variable
        self.current_route = 'main'
        self.order = 0  # flows upstream
        self.delivery = 0  # flows downstream
        self.payment = 0  # flows upstream
        self.fulfilment_rate = 1
        self.alternative_route = []
        self.alternative_route_time_cost = 0
        self.alternative_route_cost_per_ton = 0
        self.price = 1

    def calculate_fulfilment_rate(self):
        if self.order < EPSILON:
            self.fulfilment_rate = 1
        elif self.delivery > self.order + EPSILON:
            self.fulfilment_rate = 1
        else:
            self.fulfilment_rate = self.delivery / self.order

    def store_route_information(self, route: Route, main_or_alternative: str):

        if main_or_alternative == "main":
            self.route = route
            self.route_length = route.length
            # self.route_time_cost = route_time_cost
            self.route_cost_per_ton = route.cost_per_ton

        elif main_or_alternative == "alternative":
            self.alternative_route = route
            self.alternative_route_length = route.length
            # self.alternative_route_time_cost = route_time_cost
            self.alternative_route_cost_per_ton = route.cost_per_ton

        else:
            raise ValueError("'main_or_alternative' is not in ['main', 'alternative']")



from typing import TYPE_CHECKING

import networkx as nx
import numpy as np
import pandas as pd
import logging
from .caching_functions import \
    load_cached_transport_network, \
    load_cached_agent_data, \
    load_cached_transaction_table, \
    cache_transport_network, \
    cache_agent_data, load_cached_sc_network, cache_sc_network, load_cached_logistic_routes, cache_logistic_routes
from src.model.check_functions import compare_production_purchase_plans
from src.model.country_builder_functions import create_countries_from_mrio, create_countries
from src.model.firm_builder_functions import define_firms_from_local_economic_data, define_firms_from_network_data, \
    define_firms_from_mrio, create_firms, load_technical_coefficients, calibrate_input_mix, load_mrio_tech_coefs, \
    load_inventories
from src.model.household_builder_functions import define_households_from_mrio, define_households, \
    add_households_for_firms, \
    create_households
from src.model.transport_network_builder_functions import \
    create_transport_network
from src.model.builder_functions import \
    filter_sector, \
    extract_final_list_of_sector, \
    load_ton_usd_equivalence
from src.parameters import Parameters
from src.disruption.disruption import DisruptionList, TransportDisruption, CapitalDestruction
from src.simulation.simulation import Simulation
from src.network.sc_network import ScNetwork

if TYPE_CHECKING:
    from src.agents.country import Countries
    from src.agents.firm import Firms
    from src.agents.household import Households


class Model(object):
    def __init__(self, parameters: Parameters):
        # Parameters and filepath
        self.parameters = parameters
        # Initialization states
        self.transport_network_initialized = False
        self.agents_initialized = False
        self.sc_network_initialized = False
        self.logistic_routes_initialized = False
        # Main reference table
        self.sector_table = None
        # Transport network variables
        self.transport_edges = None
        self.transport_nodes = None
        self.transport_network = None
        # Agent variables
        self.firms = None
        self.firm_table = None
        self.households = None
        self.household_table = None
        self.countries = None
        self.transaction_table = None
        # Supply-chain network variables
        self.sc_network = None
        # Disruption variable
        self.disruption_list = None
        self.reconstruction_market = None

    def is_initialized(self):
        if all([self.transport_network_initialized, self.agents_initialized,
                self.sc_network_initialized, self.logistic_routes_initialized]):
            return True
        else:
            return False

    def setup_transport_network(self, cached: bool):
        if cached:
            self.transport_network, self.transport_nodes, self.transport_edges = \
                load_cached_transport_network()
        else:
            self.transport_network, self.transport_nodes, self.transport_edges = \
                create_transport_network(
                    transport_modes=self.parameters.transport_modes,
                    filepaths=self.parameters.filepaths,
                    transport_cost_data=self.parameters.transport_cost_data,
                    time_resolution=self.parameters.time_resolution
                )

            data_to_cache = {
                "transport_network": self.transport_network,
                'transport_nodes': self.transport_nodes,
                'transport_edges': self.transport_edges
            }
            cache_transport_network(data_to_cache)

        self.transport_network.define_weights(
            route_optimization_weight=self.parameters.route_optimization_weight
        )
        self.transport_network.log_km_per_transport_modes()  # Print data on km per mode
        self.transport_network_initialized = True

    def setup_firms(self):
        # TODO write
        pass

    def setup_households(self):
        pass

    def setup_countries(self):
        pass

    def setup_agents(self, cached: bool):
        if cached:
            self.sector_table, self.firms, self.firm_table, self.households, self.household_table, \
                self.countries = load_cached_agent_data()
            if self.parameters.firm_data_type == "supplier-buyer network":
                self.transaction_table = load_cached_transaction_table()
        else:
            logging.info('Filtering the sectors based on their output. ' +
                         "Cutoff type is " + self.parameters.cutoff_sector_output['type'] +
                         ", cutoff value is " + str(self.parameters.cutoff_sector_output['value']))
            self.sector_table = pd.read_csv(self.parameters.filepaths['sector_table'])
            filtered_sectors = filter_sector(self.sector_table,
                                             cutoff_sector_output=self.parameters.cutoff_sector_output,
                                             cutoff_sector_demand=self.parameters.cutoff_sector_demand,
                                             combine_sector_cutoff=self.parameters.combine_sector_cutoff,
                                             sectors_to_include=self.parameters.sectors_to_include,
                                             sectors_to_exclude=self.parameters.sectors_to_exclude)
            output_selected = self.sector_table.loc[self.sector_table['sector'].isin(filtered_sectors), 'output'].sum()
            final_demand_selected = self.sector_table.loc[
                self.sector_table['sector'].isin(filtered_sectors), 'final_demand'].sum()
            logging.info(
                str(len(filtered_sectors)) + ' sectors selected over ' + str(
                    self.sector_table.shape[0]) + ' representing ' +
                "{:.0f}%".format(output_selected / self.sector_table['output'].sum() * 100) + ' of total output and ' +
                "{:.0f}%".format(
                    final_demand_selected / self.sector_table['final_demand'].sum() * 100) + ' of final demand'
            )
            logging.info('The filtered sectors are: ' + str(filtered_sectors))

            logging.info('Generating the firms')
            if self.parameters.firm_data_type == "disaggregating IO":
                self.firm_table, firm_table_per_region = define_firms_from_local_economic_data(
                    filepath_region_economic_data=self.parameters.filepaths['region_data'],
                    sectors_to_include=filtered_sectors,
                    transport_nodes=self.transport_nodes,
                    filepath_sector_table=self.parameters.filepaths['sector_table'],
                    min_nb_firms_per_sector=self.parameters.min_nb_firms_per_sector)
            elif self.parameters.firm_data_type == "supplier-buyer network":
                self.firm_table = define_firms_from_network_data(
                    filepath_firm_table=self.parameters.filepaths['firm_table'],
                    filepath_location_table=self.parameters.filepaths['location_table'],
                    sectors_to_include=filtered_sectors,
                    transport_nodes=self.transport_nodes,
                    filepath_sector_table=self.parameters.filepaths['sector_table'])
            elif self.parameters.firm_data_type == "mrio":
                self.firm_table = define_firms_from_mrio(
                    filepath_mrio=self.parameters.filepaths['mrio'],
                    filepath_sector_table=self.parameters.filepaths['sector_table'],
                    filepath_region_table=self.parameters.filepaths['region_table'],
                    transport_nodes=self.transport_nodes,
                    io_cutoff=self.parameters.io_cutoff)
            else:
                raise ValueError(f"{self.parameters.firm_data_type} should be one of 'disaggregating', "
                                 f"'supplier-buyer network', 'mrio'")
            nb_firms = 'all'  # Weird
            logging.info('Creating firms. nb_firms: ' + str(nb_firms) +
                         ' inventory_restoration_time: ' + str(self.parameters.inventory_restoration_time) +
                         ' utilization_rate: ' + str(self.parameters.utilization_rate))
            self.firms = create_firms(
                firm_table=self.firm_table,
                keep_top_n_firms=nb_firms,
                inventory_restoration_time=self.parameters.inventory_restoration_time,
                utilization_rate=self.parameters.utilization_rate,
                capital_to_value_added_ratio=self.parameters.capital_to_value_added_ratio
            )

            n, present_sectors, flow_types_to_export = extract_final_list_of_sector(self.firms)

            # Create households
            logging.info('Defining the number of households to generate and their purchase plan')
            if self.parameters.firm_data_type == "mrio":
                self.household_table, household_sector_consumption = define_households_from_mrio(
                    filepath_mrio=self.parameters.filepaths['mrio'],
                    filepath_region_table=self.parameters.filepaths['region_table'],
                    transport_nodes=self.transport_nodes,
                    time_resolution=self.parameters.time_resolution,
                    target_units=self.parameters.monetary_units_in_model,
                    input_units=self.parameters.monetary_units_inputed
                )
            else:
                self.household_table, household_sector_consumption = define_households(
                    sector_table=self.sector_table,
                    filepath_region_data=self.parameters.filepaths['region_data'],
                    filtered_sectors=present_sectors,
                    pop_cutoff=self.parameters.pop_cutoff,
                    pop_density_cutoff=self.parameters.pop_density_cutoff,
                    local_demand_cutoff=self.parameters.local_demand_cutoff,
                    transport_nodes=self.transport_nodes,
                    time_resolution=self.parameters.time_resolution,
                    target_units=self.parameters.monetary_units_in_model,
                    input_units=self.parameters.monetary_units_inputed
                )
                cond_no_household = ~self.firm_table['od_point'].isin(self.household_table['od_point'])
                if cond_no_household.sum() > 0:
                    logging.info('We add local households for firms')
                    self.household_table, household_sector_consumption = add_households_for_firms(
                        firm_table=self.firm_table,
                        household_table=self.household_table,
                        filepath_region_data=self.parameters.filepaths['region_data'],
                        sector_table=self.sector_table,
                        filtered_sectors=present_sectors,
                        time_resolution=self.parameters.time_resolution,
                        target_units=self.parameters.monetary_units_in_model,
                        input_units=self.parameters.monetary_units_inputed
                    )
            self.households = create_households(
                household_table=self.household_table,
                household_sector_consumption=household_sector_consumption
            )

            # Loading the technical coefficients
            if self.parameters.firm_data_type == "disaggregating IO":
                import_code_in_table = self.sector_table.loc[self.sector_table['type'] == 'imports', 'sector'].iloc[
                    0]  # usually it is IMP
                load_technical_coefficients(
                    self.firms, self.parameters.filepaths['tech_coef'], self.parameters.io_cutoff,
                    import_code_in_table
                )

            elif self.parameters.firm_data_type == "supplier-buyer network":
                self.firms, self.transaction_table = calibrate_input_mix(
                    firms=self.firms,
                    firm_table=self.firm_table,
                    sector_table=self.sector_table,
                    filepath_transaction_table=self.parameters.filepaths['transaction_table']
                )

            elif self.parameters.firm_data_type == "mrio":
                load_mrio_tech_coefs(
                    firms=self.firms,
                    filepath_mrio=self.parameters.filepaths['mrio'],
                    io_cutoff=self.parameters.io_cutoff
                )

            else:
                raise ValueError(
                    f"{self.parameters.firm_data_type} should be "
                    f"one of 'disaggregating', 'supplier-buyer network', 'mrio'"
                )

            # Loading the inventories
            load_inventories(
                firms=self.firms,
                inventory_duration_target=self.parameters.inventory_duration_target,
                given_time_unit=self.parameters.inventory_duration_target_unit,
                model_time_unit=self.parameters.time_resolution,
                filepath_inventory_duration_targets=self.parameters.filepaths['inventory_duration_targets'],
                extra_inventory_target=self.parameters.extra_inventory_target,
                inputs_with_extra_inventories=self.parameters.inputs_with_extra_inventories,
                buying_sectors_with_extra_inventories=self.parameters.buying_sectors_with_extra_inventories,
                min_inventory=1
            )

            # Create agents: Countries
            if self.parameters.firm_data_type == "mrio":
                self.countries = create_countries_from_mrio(
                    filepath_mrio=self.parameters.filepaths['mrio'],
                    transport_nodes=self.transport_nodes,
                    time_resolution=self.parameters.time_resolution,
                    target_units=self.parameters.monetary_units_in_model,
                    input_units=self.parameters.monetary_units_inputed
                )
            else:
                self.countries = create_countries(
                    filepath_imports=self.parameters.filepaths['imports'],
                    filepath_exports=self.parameters.filepaths['exports'],
                    filepath_transit=self.parameters.filepaths['transit'],
                    transport_nodes=self.transport_nodes,
                    present_sectors=present_sectors,
                    countries_to_include=self.parameters.countries_to_include,
                    time_resolution=self.parameters.time_resolution,
                    target_units=self.parameters.monetary_units_in_model,
                    input_units=self.parameters.monetary_units_inputed
                )

            # Specify the weight of a unit worth of good, which may differ according to sector, or even to each
            # firm/countries Note that for imports, i.e. for the goods delivered by a country, and for transit flows,
            # we do not disentangle sectors In this case, we use an average.
            load_ton_usd_equivalence(
                sector_table=self.sector_table,
                firm_table=self.firm_table,
                firms=self.firms,
                countries=self.countries
            )

            # Save to tmp folder
            data_to_cache = {
                "sector_table": self.sector_table,
                'firm_table': self.firm_table,
                'present_sectors': present_sectors,
                'flow_types_to_export': flow_types_to_export,
                'firms': self.firms,
                'household_table': self.household_table,
                'households': self.households,
                'countries': self.countries
            }
            if self.parameters.firm_data_type == "supplier-buyer network":
                data_to_cache['transaction_table'] = self.transaction_table
            cache_agent_data(data_to_cache)

        # Locate firms and households on transport network
        self.transport_network.locate_firms_on_nodes(self.firms, self.transport_nodes)
        self.transport_network.locate_households_on_nodes(self.households, self.transport_nodes)
        self.agents_initialized = True

    def setup_sc_network(self, cached: bool):
        if cached:
            self.sc_network, self.firms, self.households, self.countries = load_cached_sc_network()

        else:
            logging.info(
                f'The supply chain graph is being created. nb_suppliers_per_input: '
                f'{self.parameters.nb_suppliers_per_input}')
            self.sc_network = ScNetwork()

            logging.info('Households are selecting their retailers (domestic B2C flows and import B2C flows)')
            for household in self.households.values():
                household.select_suppliers(self.sc_network, self.firms, self.countries,
                                           self.parameters.nb_suppliers_per_input, self.parameters.force_local_retailer,
                                           self.parameters.weight_localization_household,
                                           self.parameters.firm_data_type)

            logging.info('Exporters are being selected by purchasing countries (export B2B flows)')
            logging.info('and trading countries are being connected (transit flows)')
            for country in self.countries.values():
                country.select_suppliers(self.sc_network, self.firms, self.countries,
                                         self.sector_table, self.transport_nodes)

            logging.info(
                f'Firms are selecting their domestic and international suppliers (import B2B flows) '
                f'(domestic B2B flows). Weight localisation is {self.parameters.weight_localization_firm}'
            )
            import_code_from_table = self.sector_table.loc[self.sector_table['type'] == 'imports', 'sector'].iloc[0]

            if self.parameters.firm_data_type in ["disaggregating IO", 'mrio']:
                for firm in self.firms.values():
                    firm.select_suppliers(self.sc_network, self.firms, self.countries,
                                          self.parameters.nb_suppliers_per_input,
                                          self.parameters.weight_localization_firm,
                                          self.parameters.firm_data_type,
                                          import_code=import_code_from_table)

            elif self.parameters.firm_data_type == "supplier-buyer network":
                for firm in self.firms.values():
                    inputed_supplier_links = self.transaction_table[self.transaction_table['buyer_id'] == firm.pid]
                    output = self.firm_table.set_index('id').loc[firm.pid, "output"]
                    firm.select_suppliers_from_data(self.sc_network, self.firms, self.countries,
                                                    inputed_supplier_links, output,
                                                    import_code=import_code_from_table)

            else:
                raise ValueError(self.parameters.firm_data_type +
                                 " should be one of 'disaggregating IO', 'supplier-buyer network', 'mrio'")

            firm_pids = list(self.firms.keys())
            node_pid_in_sc_network = [node.pid for node in self.sc_network]
            if len(set(firm_pids) - set(node_pid_in_sc_network)) > 0:
                unconnected_firms = list(set(firm_pids) - set(node_pid_in_sc_network))
                for firm_pid in unconnected_firms:
                    print(self.firms[firm_pid].id_str())
                raise ValueError('Some firms are not in the sc network')

            self.sc_network.remove_useless_commercial_links()

            logging.info('The nodes and edges of the supplier--buyer have been created')
            # Save to tmp folder
            data_to_cache = {
                "supply_chain_network": self.sc_network,
                'firms': self.firms,
                'households': self.households,
                'countries': self.countries
            }
            cache_sc_network(data_to_cache)

            self.sc_network_initialized = True

    def setup_logistic_routes(self, cached: bool):
        if cached:
            self.sc_network, self.transport_network, self.firms, self.households, \
                self.countries = load_cached_logistic_routes()

        else:
            logging.info('The supplier--buyer graph is being connected to the transport network')
            logging.info('Each B2B and transit edge is being linked to a route of the transport network')
            logging.info('Routes for transit and import flows are being selected by trading countries')
            for country in self.countries.values():
                country.choose_initial_routes(self.sc_network, self.transport_network,
                                              self.parameters.capacity_constraint,
                                              self.parameters.transport_cost_noise_level,
                                              self.parameters.monetary_units_in_model)
            logging.info('Routes for exports and B2B domestic flows are being selected by domestic firms')
            for firm in self.firms.values():
                if firm.sector_type not in self.parameters.sectors_no_transport_network:
                    firm.choose_initial_routes(self.sc_network, self.transport_network,
                                               self.parameters.capacity_constraint,
                                               self.parameters.transport_cost_noise_level,
                                               self.parameters.monetary_units_in_model)
            # Save to tmp folder
            data_to_cache = {
                'transport_network': self.transport_network,
                "supply_chain_network": self.sc_network,
                'firms': self.firms,
                'households': self.households,
                'countries': self.countries
            }
            cache_logistic_routes(data_to_cache)

            self.logistic_routes_initialized = True

    def reset_variables(self):
        logging.info("Resetting variables on transport network")
        self.transport_network.reinitialize_flows_and_disruptions()

        logging.info("Resetting agents and commercial links variables")
        for household in self.households.values():
            household.reset_variables()
            for edge in self.sc_network.in_edges(household):
                self.sc_network[edge[0]][household]['object'].reset_variables()
        for firm in self.firms.values():
            firm.reset_variables()
            for edge in self.sc_network.in_edges(firm):
                self.sc_network[edge[0]][firm]['object'].reset_variables()
        for country in self.countries.values():
            country.reset_variables()
            for edge in self.sc_network.in_edges(country):
                self.sc_network[edge[0]][country]['object'].reset_variables()

    def set_initial_conditions(self):
        logging.info("Setting initial conditions to input-output equilibrium")
        """
        Initialize the supply chain network at the input--output equilibrium
    
        We will use the matrix forms to solve the following equation for X (production):
        D + E + AX = X + I
        where:
            D: final demand from households
            E: exports
            I: imports
            X: firm productions
            A: the input-output matrix
        These vectors and matrices are in the firm-and-country space.
        """

        # Get the weighted connectivity matrix.
        # Weight is the sectoral technical coefficient, if there is only one supplier for the input
        # It there are several, the technical coefficient is multiplied by the share of input of
        # this type that the firm buys to this supplier.
        # l1 = [firm.pid for firm in self.sc_network.nodes if isinstance(firm.pid, int) ]
        # l1.sort()
        # print(l1)
        # print([firm.pid for firm in self.firms])
        firm_connectivity_matrix = nx.adjacency_matrix(
            self.sc_network,
            # graph.subgraph(list(graph.nodes)[:-1]),
            weight='weight',
            nodelist=self.firms.values()
        ).todense()
        # Imports are considered as "a sector". We get the weight per firm for these inputs.
        # TODO !!! aren't I computing the same thing as the IMP tech coef? To check
        import_weight_per_firm = [
            sum([
                self.sc_network[supply_edge[0]][supply_edge[1]]['weight']
                for supply_edge in self.sc_network.in_edges(firm)
                if self.sc_network[supply_edge[0]][supply_edge[1]]['object'].category == 'import'
            ])
            for firm in self.firms.values()
        ]
        n = len(self.firms)

        # Build final demand vector per firm, of length n
        # Exports are considered as final demand
        final_demand_vector = self.build_final_demand_vector(self.households, self.countries, self.firms)

        # Solve the input--output equation
        eq_production_vector = np.linalg.solve(
            np.eye(n) - firm_connectivity_matrix,
            final_demand_vector  # + 0.01
        )

        # Initialize households variables
        for household in self.households.values():
            household.initialize_var_on_purchase_plan()

        # Compute costs
        # 1. Input costs
        domestic_input_cost_vector = np.multiply(
            firm_connectivity_matrix.sum(axis=0).reshape((n, 1)),
            eq_production_vector
        )
        import_input_cost_vector = np.multiply(
            np.array(import_weight_per_firm).reshape((n, 1)),
            eq_production_vector
        )
        input_cost_vector = domestic_input_cost_vector + import_input_cost_vector
        # 2. Transport costs
        proportion_of_transport_cost_vector = 0.2 * np.ones((n, 1))  # TODO should be parametrized
        transport_cost_vector = np.multiply(eq_production_vector, proportion_of_transport_cost_vector)
        # 3. Compute other costs based on margin
        margin = np.array([firm.target_margin for firm in self.firms.values()]).reshape((n, 1))
        other_cost_vector = np.multiply(eq_production_vector, (1 - margin)) - input_cost_vector - transport_cost_vector

        # Based on these calculus, update agents variables
        # 1. Firm operational variables
        for firm in self.firms.values():  # TODO make it a FirmCollection method
            firm.initialize_operational_variables(
                eq_production=eq_production_vector[(firm.pid, 0)]
            )
        # 2. Firm financial variables
        for firm in self.firms.values():
            firm.initialize_financial_variables(
                eq_production=eq_production_vector[(firm.pid, 0)],
                eq_input_cost=input_cost_vector[(firm.pid, 0)],
                eq_transport_cost=transport_cost_vector[(firm.pid, 0)],
                eq_other_cost=other_cost_vector[(firm.pid, 0)]
            )
        # 3. Commercial links: agents set their order
        for household in self.households.values():
            household.send_purchase_orders(self.sc_network)
        for country in self.countries.values():
            country.send_purchase_orders(self.sc_network)
        for firm in self.firms.values():
            firm.send_purchase_orders(self.sc_network)
        # 4. The following is just to set once for all the share of sales of each client
        for firm in self.firms.values():
            firm.retrieve_orders(self.sc_network)
            firm.aggregate_orders(log_info=True)
            firm.eq_total_order = firm.total_order
            firm.calculate_client_share_in_sales()

        # Set price to 1
        self.reset_prices()

    def reset_prices(self):
        # set prices to 1
        for edge in self.sc_network.edges:
            self.sc_network[edge[0]][edge[1]]['object'].price = 1

    @staticmethod
    def build_final_demand_vector(households: "Households", countries: "Countries", firms: "Firms") -> np.array:
        """
        Create a numpy.Array of the final demand per firm, including exports

        Households and countries should already have set their purchase plan

        Returns
        -------
        numpy.Array of dimension (len(firms), 1)
        """
        final_demand_vector = np.zeros((len(firms), 1))

        # Collect households final demand. They buy only from firms.
        for household in households.values():
            for retailer_id, quantity in household.purchase_plan.items():
                if isinstance(retailer_id, int):  # we only consider purchase from firms, not from other countries
                    final_demand_vector[(retailer_id, 0)] += quantity

        # Collect country final demand. They buy from firms and countries.
        # We need to filter the demand directed to firms only.
        for country in countries.values():
            for supplier_id, quantity in country.purchase_plan.items():
                if isinstance(supplier_id, int):  # we only consider purchase from firms, not from other countries
                    final_demand_vector[(supplier_id, 0)] += quantity

        return final_demand_vector

    def run_static(self):
        simulation = Simulation("initial_state")
        logging.info("Simulating the initial state")
        self.run_one_time_step(time_step=0, current_simulation=simulation)
        return simulation

    def run_disruption(self):
        # Initialize the model
        simulation = Simulation("event")
        logging.info("Simulating the initial state")
        self.run_one_time_step(time_step=0, current_simulation=simulation)

        # Get disruptions
        self.disruption_list = DisruptionList.from_events_parameter(self.parameters.events,
                                                                    self.parameters.monetary_units_in_model,
                                                                    self.transport_edges, self.firm_table,
                                                                    self.firms)
        if len(self.disruption_list) == 0:
            raise ValueError("No disruption could be read")
        logging.info(f"{len(self.disruption_list)} disruption(s) will occur")
        self.disruption_list.log_info()

        # Adjust t_final
        t_final = self.parameters.duration_dic[self.disruption_list.end_time]
        logging.info('Simulation will last at max ' + str(t_final) + ' time steps.')

        logging.info("Starting time loop")
        for t in range(1, t_final + 1):
            logging.info(f'Time t={t}')
            self.run_one_time_step(time_step=t, current_simulation=simulation)

            if (t > max([disruption.start_time for disruption in
                         self.disruption_list])) and self.parameters.epsilon_stop_condition:
                if self.is_back_to_equilibrium:
                    logging.info("Simulation stops")
                    break
        return simulation

    def run_one_time_step(self, time_step: int, current_simulation: Simulation):
        self.transport_network.reset_current_loads(self.parameters.route_optimization_weight)

        if self.disruption_list:
            self.apply_disruption(time_step)

        self.firms.retrieve_orders(self.sc_network)
        if self.reconstruction_market:
            self.reconstruction_market.evaluate_demand_to_firm(self.firms)
            self.reconstruction_market.send_orders(self.firms)
        self.firms.plan_production(self.sc_network, self.parameters.propagate_input_price_change)
        self.firms.plan_purchase(self.parameters.adaptive_inventories, self.parameters.adaptive_supplier_weight)
        self.households.send_purchase_orders(self.sc_network)
        self.countries.send_purchase_orders(self.sc_network)
        self.firms.send_purchase_orders(self.sc_network)
        self.firms.produce()
        self.countries.deliver(self.sc_network, self.transport_network, self.parameters.sectors_no_transport_network,
                               self.parameters.rationing_mode, self.parameters.capacity_constraint,
                               self.parameters.monetary_units_in_model, self.parameters.cost_repercussion_mode,
                               self.parameters.price_increase_threshold, self.parameters.transport_cost_noise_level)
        self.firms.deliver(self.sc_network, self.transport_network, self.parameters.sectors_no_transport_network,
                           self.parameters.rationing_mode, self.parameters.capacity_constraint,
                           self.parameters.monetary_units_in_model, self.parameters.cost_repercussion_mode,
                           self.parameters.price_increase_threshold, self.parameters.transport_cost_noise_level)
        if self.reconstruction_market:
            self.reconstruction_market.distribute_new_capital(self.firms)
        # if congestion: TODO reevaluate modeling of congestion
        #     if (time_step == 0):
        #         transport_network.evaluate_normal_traffic()
        #     else:
        #         transport_network.evaluate_congestion()
        #         if len(transport_network.congestionned_edges) > 0:
        #             logging.info("Nb of congestionned segments: " +
        #                          str(len(transport_network.congestionned_edges)))
        #     for firm in firms:
        #         firm.add_congestion_malus2(sc_network, transport_network)
        #     for country in countries:
        #         country.add_congestion_malus2(sc_network, transport_network)
        #
        if time_step in [0, 1]:
            current_simulation.transport_network_data += self.transport_network.compute_flow_per_segment(time_step)
        # TODO: store transport data, depending on current_simulation type and time step
        # TODO: store supply chain data, depending on current_simulation type and time step
        # if (time_step in [0, 1, 2]) and (
        # export_flows):  # should be done at this stage, while the goods are on their way
        #     collect_shipments = False
        #     transport_network.compute_flow_per_segment(flow_types_to_export)
        #     observer.collect_transport_flows(transport_network,
        #                                      time_step=time_step, flow_types=flow_types_to_export,
        #                                      collect_shipments=collect_shipments)
        #     exportTransportFlows(observer, export_folder)
        #     exportTransportFlowsLayer(observer, export_folder, time_step=time_step,
        #                               transport_edges=transport_edges)
        #     if sum([len(sublist) for sublist in observer.specific_edges_to_monitor.values()]) > 0:
        #         observer.collect_specific_flows(transport_network)
        #         exportSpecificFlows(observer, export_folder)
        #     if collect_shipments:
        #         exportShipmentsLayer(observer, export_folder, time_step=time_step,
        #                              transport_edges=transport_edges)
        #
        # if (time_step == 0) and (
        # export_sc_flow_analysis):  # should be done at this stage, while the goods are on their way
        #     analyzeSupplyChainFlows(sc_network, firms, export_folder)
        #
        self.households.receive_products(self.sc_network, self.transport_network,
                                         self.parameters.sectors_no_transport_network)
        self.countries.receive_products(self.sc_network, self.transport_network,
                                        self.parameters.sectors_no_transport_network)
        self.firms.receive_products(self.sc_network, self.transport_network,
                                    self.parameters.sectors_no_transport_network)
        self.firms.evaluate_profit(self.sc_network)

        self.transport_network.update_road_disruption_state()
        self.firms.update_disrupted_production_capacity()
        #
        self.store_agent_data(time_step, current_simulation)

        compare_production_purchase_plans(self.firms, self.countries, self.households)

    def apply_disruption(self, time_step: int):
        disruptions_starting_now = self.disruption_list.filter_start_time(time_step)
        for disruption in disruptions_starting_now:
            if isinstance(disruption, TransportDisruption):
                disruption.implement(self.transport_network)
            if isinstance(disruption, CapitalDestruction):
                disruption.implement(self.firms, self)
        # edge_disruptions_starting_now = disruptions_starting_now.filter_type('transport_edge')
        # if len(edge_disruptions_starting_now) > 0:
        #     self.transport_network.disrupt_edges(
        #         edge_disruptions_starting_now.get_item_id_duration_reduction_dict()
        #     )
        # firm_disruptions_starting_now = disruptions_starting_now.filter_type('firm')
        # if len(firm_disruptions_starting_now) > 0:
        #     self.firms.get_disrupted(firm_disruptions_starting_now.get_item_id_duration_reduction_dict())
        # node disruption not implemented

    @property
    def is_back_to_equilibrium(self):
        household_extra_spending = sum([household.extra_spending for household in self.households.values()])
        household_consumption_loss = sum([household.consumption_loss for household in self.households.values()])
        country_extra_spending = sum([country.extra_spending for country in self.countries.values()])
        country_consumption_loss = sum([country.consumption_loss for country in self.countries.values()])
        if (household_extra_spending <= self.parameters.epsilon_stop_condition) & \
                (household_consumption_loss <= self.parameters.epsilon_stop_condition) & \
                (country_extra_spending <= self.parameters.epsilon_stop_condition) & \
                (country_consumption_loss <= self.parameters.epsilon_stop_condition):
            logging.info('Household and country extra spending and consumption loss are at pre-disruption values.')
            return True
        else:
            return False

    def store_agent_data(self, time_step: int, simulation: Simulation):
        # TODO: could create agent-level method to export stuff
        simulation.firm_data += [
            {
                'time_step': time_step,
                'firm': firm.pid,
                'production': firm.production,
                'profit': firm.profit,
                'transport_cost': firm.finance['costs']['transport'],
                'input_cost': firm.finance['costs']['input'],
                'other_cost': firm.finance['costs']['other'],
                'inventory_duration': firm.current_inventory_duration,
                'generalized_transport_cost': firm.generalized_transport_cost,
                'usd_transported': firm.usd_transported,
                'tons_transported': firm.tons_transported,
                'tonkm_transported': firm.tonkm_transported
            }
            for firm in self.firms.values()
        ]
        simulation.country_data += [
            {
                'time_step': time_step,
                'country': country.pid,
                'generalized_transport_cost': country.generalized_transport_cost,
                'usd_transported': country.usd_transported,
                'tons_transported': country.tons_transported,
                'tonkm_transported': country.tonkm_transported,
                'extra_spending': country.extra_spending,
                'consumption_loss': country.consumption_loss,
                'spending': sum(list(country.qty_purchased.values()))
            }
            for country in self.countries.values()
        ]
        simulation.household_data += [
            {
                'time_step': time_step,
                'household': household.pid,
                'spending_per_retailer': household.spending_per_retailer,
                'consumption_per_retailer': household.consumption_per_retailer,
                'extra_spending_per_sector': household.extra_spending_per_sector,
                'consumption_loss_per_sector': household.consumption_loss_per_sector,
                'extra_spending': household.extra_spending,
                'consumption_loss': household.consumption_loss
            }
            for household in self.households.values()
        ]

    def export_transport_nodes_edges(self):
        self.transport_nodes.to_file(
            self.parameters.export_folder / 'transport_nodes.geojson',
            driver="GeoJSON", index=False)
        self.transport_edges.to_file(
            self.parameters.export_folder / 'transport_edges.geojson',
            driver="GeoJSON", index=False)

    def export_agent_tables(self):
        self.firm_table.to_csv(self.parameters.export_folder / 'firm_table.csv', index=False)
        self.household_table.to_csv(self.parameters.export_folder / 'household_table.csv', index=False)



import logging
import math

import numpy as np
import pandas as pd


def generate_weights(nb_suppliers: int, importance_of_each: list or None):
    # if there is only one supplier, return 1
    if nb_suppliers == 1:
        return [1]

    # if there are several and importance are provided, choose according to importance
    if importance_of_each:
        return [x / sum(importance_of_each) for x in importance_of_each]

    # otherwise choose random values
    else:
        rdm_values = np.random.uniform(0, 1, size=nb_suppliers)
        return list(rdm_values / sum(rdm_values))


def generate_weights_from_list(list_nb):
    sum_list = sum(list_nb)
    return [nb / sum_list for nb in list_nb]


def rescale_values(input_list, minimum=0.1, maximum=1, max_val=None, alpha=1, normalize=False):
    max_val = max_val or max(input_list)
    min_val = min(input_list)
    if max_val == min_val:
        res = [0.5 * maximum] * len(input_list)
    else:
        res = [
            minimum + (((val - min_val) / (max_val - min_val)) ** alpha) * (maximum - minimum)
            for val in input_list
        ]
    if normalize:
        res = [x / sum(res) for x in res]
    return res


def calculate_distance_between_agents(agentA, agentB):
    if (agentA.od_point == -1) or (agentB.od_point == -1):
        logging.warning("Try to calculate distance between agents, but one of them does not have real od point")
        return 1
    else:
        return compute_distance_from_arcmin(agentA.long, agentA.lat, agentB.long, agentB.lat)


def compute_distance_from_arcmin(x0, y0, x1, y1):
    # This is a very approximate way to convert arc distance into km
    EW_dist = (x1 - x0) * 112.5
    NS_dist = (y1 - y0) * 111
    return math.sqrt(EW_dist ** 2 + NS_dist ** 2)


def add_or_increment_dict_key(dic: dict, key, value: float | int):
    if key not in dic.keys():
        dic[key] = value
    else:
        dic[key] += value


def rescale_monetary_values(
        values: pd.Series | pd.DataFrame | float,
        time_resolution: str = "week",
        target_units: str = "mUSD",
        input_units: str = "USD"
) -> pd.Series | pd.DataFrame | float:
    """Rescale monetary values using the appropriate timescale and monetary units

    Parameters
    ----------
    values : pandas.Series, pandas.DataFrame, float
        Values to transform

    time_resolution : 'day', 'week', 'month', 'year'
        The number in the input table are yearly figure

    target_units : 'USD', 'kUSD', 'mUSD'
        Monetary units to which values are converted

    input_units : 'USD', 'kUSD', 'mUSD'
        Monetary units of the inputted values

    Returns
    -------
    same type as values
    """
    # Rescale according to the time period chosen
    periods = {'day': 365, 'week': 52, 'month': 12, 'year': 1}
    values = values / periods[time_resolution]

    # Change units
    units = {"USD": 1, "kUSD": 1e3, "mUSD": 1e6}
    values = values * units[input_units] / units[target_units]

    return values


def add_or_append_to_dict(dictionary, key, value_to_add):
    if key in dictionary.keys():
        dictionary[key] += value_to_add
    else:
        dictionary[key] = value_to_add



import logging

import pandas as pd


def compare_production_purchase_plans(firms, countries, households):
    # Create dictionary to map firm id and country id into sector
    dic_agent_id_to_sector = {pid: firm.sector for pid, firm in firms.items()}
    for pid, country in countries.items():
        dic_agent_id_to_sector[pid] = "IMP"

    # Evaluate purchase plans of firms
    df = pd.DataFrame({pid: firm.purchase_plan for pid, firm in firms.items()})
    df["tot_purchase_planned_by_firms"] = df.sum(axis=1)
    df['input_sector'] = df.index.map(dic_agent_id_to_sector)
    df_firms = df.groupby('input_sector')["tot_purchase_planned_by_firms"].sum()

    # of countries
    if len(countries) == 1:
        country_id = list(countries.keys())[0]
        df = pd.DataFrame(pd.Series(countries[country_id].purchase_plan, name=country_id))
    df = pd.DataFrame({pid: country.purchase_plan for pid, country in countries.items()})
    df["tot_purchase_planned_by_countries"] = df.sum(axis=1)
    df['input_sector'] = df.index.map(dic_agent_id_to_sector)
    df_countries = df.groupby('input_sector')["tot_purchase_planned_by_countries"].sum()
    # of households
    df = pd.DataFrame({pid: household.purchase_plan for pid, household in households.items()})
    df["tot_purchase_planned_by_households"] = df.sum(axis=1)
    # df = pd.DataFrame({"tot_purchase_planned_by_households": households.purchase_plan})
    df['input_sector'] = df.index.map(dic_agent_id_to_sector)
    df_households = df.groupby('input_sector')["tot_purchase_planned_by_households"].sum()

    # concat
    df_purchase_plan = pd.concat([df_firms, df_countries, df_households], axis=1, sort=True)

    # Evalute productions/sales
    # of firms
    df = pd.DataFrame({
        "tot_production_per_firm": {pid: firm.production for pid, firm in firms.items()}
    }
    )
    df['sector'] = df.index.map(dic_agent_id_to_sector)
    df_firms = df.groupby('sector')["tot_production_per_firm"].sum()

    # of countries
    df = pd.DataFrame(
        {
            "tot_production_per_country":
                {pid: country.qty_sold for pid, country in countries.items()}
        }
    )
    df['sector'] = df.index.map(dic_agent_id_to_sector)
    df_countries = df.groupby('sector')["tot_production_per_country"].sum()

    # concat
    df_sales = pd.concat([df_firms, df_countries], axis=1, sort=True)

    # Compare
    res = pd.concat([df_purchase_plan, df_sales], axis=1, sort=True)
    res['dif'] = (res["tot_purchase_planned_by_firms"]
                  + res["tot_purchase_planned_by_countries"]
                  + res["tot_purchase_planned_by_households"]
                  - res['tot_production_per_firm'] - res['tot_production_per_country'])
    boolindex_unbalanced = res['dif'] > 1e-6
    if boolindex_unbalanced.sum() > 0:
        logging.warning("Sales does not equate purchases for sectors: " +
                        str(res.index[boolindex_unbalanced].tolist()))


# def compareDeliveredVsReceived(firm_list=None, households=None, G=None):
#     # not finished
#     qty_delivered_by_firm_per_sector = {}
#     for firm in firm_list:
#         if firm.sector not in qty_delivered_by_firm_per_sector.keys():
#             qty_delivered_by_firm_per_sector[firm.sector] = 0
#
#         for edge in G.out_edges(firm):
#             qty_delivered_by_firm_per_sector[firm.sector] += \
#                 G[firm][edge[1]]['object'].delivery
#
#         qty_bought_by_household_per_sector = {}
#         for edge in G.in_edges(households):
#             if edge[0].sector not in qty_bought_by_household_per_sector.keys():
#                 qty_bought_by_household_per_sector[edge[0].sector] = 0
#             qty_bought_by_household_per_sector[firm.sector] += \
#                 G[edge[0]][households]['object'].delivery
#
#     qty_ordered_by_firm_per_sector = {}
#     for firm in firm_list:
#         if firm.sector not in qty_ordered_by_firm_per_sector.keys():
#             qty_delivered_by_firm_per_sector[firm.sector] = 0



from typing import TYPE_CHECKING

import geopandas
import networkx as nx
import numpy as np
import pandas as pd
import logging

from src.model.basic_functions import add_or_append_to_dict
from src.network.route import Route

if TYPE_CHECKING:
    from src.network.commercial_link import CommercialLink


class TransportNetwork(nx.Graph):

    def add_transport_node(self, node_id, all_nodes_data):  # used in add_transport_edge_with_nodes
        node_attributes = ["id", "geometry", "special", "name"]
        node_data = all_nodes_data.loc[node_id, node_attributes].to_dict()
        node_data['shipments'] = {}
        node_data['disruption_duration'] = 0
        node_data['firms_there'] = []
        node_data['households_there'] = None
        node_data['type'] = 'road'
        self.add_node(node_id, **node_data)

    def log_km_per_transport_modes(self):
        km_per_mode = pd.DataFrame({
            "km": nx.get_edge_attributes(self, "km"),
            "type": nx.get_edge_attributes(self, "type")
        })
        km_per_mode = km_per_mode.groupby('type')['km'].sum().to_dict()
        logging.info("Total length of transport network is: " +
                     "{:.0f} km".format(sum(km_per_mode.values())))
        for mode, km in km_per_mode.items():
            logging.info(mode + ": {:.0f} km".format(km))
        logging.info('Nb of nodes: ' + str(len(self.nodes)) + ', Nb of edges: ' + str(len(self.edges)))

    def add_transport_edge_with_nodes(self, edge_id: int,
                                      all_edges_data: geopandas.GeoDataFrame,
                                      all_nodes_data: geopandas.GeoDataFrame):
        # Selecting data
        edge_attributes = ['id', "type", 'surface', "geometry", "class", "km", 'special', "name",
                           "capacity", "disruption",
                           "cost_per_ton", "travel_time", "time_cost", 'cost_travel_time', 'cost_variability',
                           'agg_cost']
        if all_edges_data['type'].nunique() > 1:  # if there are multiple modes
            edge_attributes += ['multimodes']
        edge_data = all_edges_data.loc[edge_id, edge_attributes].to_dict()
        end_ids = all_edges_data.loc[edge_id, ["end1", "end2"]].tolist()
        # Creating the start and end nodes
        if end_ids[0] not in self.nodes:
            self.add_transport_node(end_ids[0], all_nodes_data)
        if end_ids[1] not in self.nodes:
            self.add_transport_node(end_ids[1], all_nodes_data)
        # Creating the edge
        self.add_edge(end_ids[0], end_ids[1], **edge_data)
        # print("edge id:", edge_id, "| end1:", end_ids[0], "| end2:", end_ids[1], "| nb edges:", len(self.edges))
        # print(self.edges)
        self[end_ids[0]][end_ids[1]]['node_tuple'] = (end_ids[0], end_ids[1])
        self[end_ids[0]][end_ids[1]]['shipments'] = {}
        self[end_ids[0]][end_ids[1]]['disruption_duration'] = 0
        self[end_ids[0]][end_ids[1]]['current_load'] = 0
        self[end_ids[0]][end_ids[1]]['overused'] = False
        self[end_ids[0]][end_ids[1]]['current_capacity'] = self[end_ids[0]][end_ids[1]]['capacity']

    def define_weights(self, route_optimization_weight):
        logging.debug('Transport network: defining weights that will be used for shortest-path algorithm')
        for edge in self.edges:
            self[edge[0]][edge[1]]['weight'] = self[edge[0]][edge[1]][route_optimization_weight]
            self[edge[0]][edge[1]]['capacity_weight'] = self[edge[0]][edge[1]][route_optimization_weight]

    def locate_firms_on_nodes(self, firms, transport_nodes):
        """The nodes of the transport network stores the list of firms located there
        using the attribute "firms_there".
        There can be several firms in one node.
        "transport_nodes" is a geodataframe of the nodes. It also contains this list in the columns
        "firm_there" as a comma-separated string

        This function reinitialize those fields and repopulate them with the adequate information
        """
        # Reinitialize
        transport_nodes['firms_there'] = ""
        for node_id in self.nodes:
            self._node[node_id]['firms_there'] = []
        # Locate firms
        for firm in firms.values():
            self._node[firm.od_point]['firms_there'].append(firm.pid)
            transport_nodes.loc[transport_nodes['id'] == firm.od_point, "firms_there"] += (',' + str(firm.pid))

    def locate_households_on_nodes(self, households, transport_nodes):
        """The nodes of the transport network stores the list of households located there
        using the attribute "household_there".
        There can only be one household in one node.
        "transport_nodes" is a geodataframe of the nodes. It also contains the id of the household.

        This function reinitialize those fields and repopulate them with the adequate information
        """
        # Reinitialize
        transport_nodes['household_there'] = None
        for pid, household in households.items():
            self._node[household.od_point]['household_there'] = pid
            transport_nodes.loc[transport_nodes['id'] == household.od_point, "household_there"] = pid

    def provide_shortest_route(self, origin_node: int, destination_node: int,
                               route_weight: str, noise_level: float = 0.0) -> Route or None:
        """nx.shortest_path returns path as list of nodes
        we transform it into a route, which contains nodes and edges:
        [(1,), (1,5), (5,), (5,8), (8,)]
        """
        if origin_node not in self.nodes:
            logging.info("Origin node " + str(origin_node) + " not in the available transport network")
            return None

        elif destination_node not in self.nodes:
            logging.info("Destination node " + str(destination_node) + " not in the available transport network")
            return None

        elif nx.has_path(self, origin_node, destination_node):
            if noise_level > 0:
                self.add_noise_to_weight(route_weight, noise_level)
                sp = nx.shortest_path(self, origin_node, destination_node, weight=route_weight + '_noise')
            else:
                sp = nx.shortest_path(self, origin_node, destination_node, weight=route_weight)
            route = Route(sp, self)
            return route

        else:
            logging.info("There is no path between " + str(origin_node) + " and " + str(destination_node))
            return None

    def add_noise_to_weight(self, weight: str, noise_sd: float):
        noise_levels = np.random.normal(0, noise_sd, len(self.edges)).tolist()
        for edge in self.edges:
            self[edge[0]][edge[1]][weight + '_noise'] = self[edge[0]][edge[1]][weight] * (1 + noise_levels.pop())

    def get_undisrupted_network(self):
        available_nodes = [node for node in self.nodes if self._node[node]['disruption_duration'] == 0]
        available_subgraph = self.subgraph(available_nodes)
        available_edges = [edge for edge in self.edges if self[edge[0]][edge[1]]['disruption_duration'] == 0]
        available_subgraph = available_subgraph.edge_subgraph(available_edges)
        return TransportNetwork(available_subgraph)

    def disrupt_roads(self, disruption):
        # Disrupting nodes
        for node_id in disruption['node']:
            logging.info('Road node ' + str(node_id) +
                         ' gets disrupted for ' + str(disruption['duration']) + ' time steps')
            self._node[node_id]['disruption_duration'] = disruption['duration']
        # Disrupting edges
        for edge in self.edges:
            if self[edge[0]][edge[1]]['type'] == 'virtual':
                continue
            else:
                if self[edge[0]][edge[1]]['id'] in disruption['edge']:
                    logging.info('Road edge ' + str(self[edge[0]][edge[1]]['id']) +
                                 ' gets disrupted for ' + str(disruption['duration']) + ' time steps')
                    self[edge[0]][edge[1]]['disruption_duration'] = disruption['duration']

    def disrupt_one_edge(self, edge, capacity_reduction: float, duration: int):
        logging.info(f"Road edge {self[edge[0]][edge[1]]['id']} gets disrupted for {duration} time steps, "
                     f"capacity reduction is {capacity_reduction}")
        self[edge[0]][edge[1]]['disruption_duration'] = capacity_reduction

    def update_road_disruption_state(self):
        """
        One time step is gone
        The remaining duration of disruption is decreased by 1
        """
        for node in self.nodes:
            if self._node[node]['disruption_duration'] > 0:
                self._node[node]['disruption_duration'] -= 1
        for edge in self.edges:
            if self[edge[0]][edge[1]]['disruption_duration'] > 0:
                self[edge[0]][edge[1]]['disruption_duration'] -= 1

    def transport_shipment(self, commercial_link: "CommercialLink", capacity_constraint: bool):
        # Select the route to transport the shipment: main or alternative
        if commercial_link.current_route == 'main':
            route_to_take = commercial_link.route
        elif commercial_link.current_route == 'alternative':
            route_to_take = commercial_link.alternative_route
        else:
            route_to_take = []

        # Propagate the shipments
        for route_segment in route_to_take:
            if len(route_segment) == 2:  # pass shipments to edges
                self[route_segment[0]][route_segment[1]]['shipments'][commercial_link.pid] = {
                    "from": commercial_link.supplier_id,
                    "to": commercial_link.buyer_id,
                    "quantity": commercial_link.delivery,
                    "tons": commercial_link.delivery_in_tons,
                    "product_type": commercial_link.product_type,
                    "flow_category": commercial_link.category,
                    "price": commercial_link.price
                }
            elif len(route_segment) == 1:  # pass shipments to nodes
                self._node[route_segment[0]]['shipments'][commercial_link.pid] = {
                    "from": commercial_link.supplier_id,
                    "to": commercial_link.buyer_id,
                    "quantity": commercial_link.delivery,
                    "tons": commercial_link.delivery_in_tons,
                    "product_type": commercial_link.product_type,
                    "flow_category": commercial_link.category,
                    "price": commercial_link.price
                }

        # Propagate the load
        self.update_load_on_route(route_to_take, commercial_link.delivery_in_tons, capacity_constraint)

    def access_edge(self, edge):
        return self[edge[0]][edge[1]]

    def update_load_on_route(self, route: "Route", load: float, capacity_constraint: bool):
        """Affect a load to a route

        The current_load attribute of each edge in the route will be increased by the new load.
        A load is typically expressed in tons. If the current_load exceeds the capacity,
        then capacity_burden is added to the capacity_weight. This will prevent firms from choosing this route
        """
        # logging.info("Edge (2610, 2589): current_load "+str(self[2610][2589]['current_load']))
        capacity_burden = 1e10
        for edge in route.transport_edges:
            # Check if the edge to be used is not over capacity already
            if capacity_constraint:
                if self[edge[0]][edge[1]]['overused']:
                    logging.info(f"Edge {edge} ({self[edge[0]][edge[1]]['type']}) is over capacity and got selected")
            # Add the load
            self[edge[0]][edge[1]]['current_load'] += load
            # If it exceeds capacity, add the capacity_burden to both the mode_weight and the capacity_weight
            if capacity_constraint:
                if ~self[edge[0]][edge[1]]['overused'] and \
                        (self[edge[0]][edge[1]]['current_load'] > self[edge[0]][edge[1]]['capacity']):
                    logging.info(f"Edge {edge} ({self[edge[0]][edge[1]]['type']}) "
                                 f"has exceeded its capacity. Current load is {self[edge[0]][edge[1]]['current_load']},"
                                 f" capacity is ({self[edge[0]][edge[1]]['capacity']})")
                    self[edge[0]][edge[1]]['overused'] = True
                    self[edge[0]][edge[1]]["capacity_weight"] += capacity_burden

    def reset_current_loads(self, route_optimization_weight):
        """
        Reset current_load to 0
        If an edge was burdened due to capacity exceed, we remove the burden
        """
        for edge in self.edges:
            self[edge[0]][edge[1]]['current_load'] = 0
            self[edge[0]][edge[1]]['overused'] = False

        self.define_weights(route_optimization_weight)

    def give_route_mode(self, route):
        """
        Which mode is used on the route?
        Return the list of transport mode used along the route
        """
        modes = []
        all_edges = [item for item in route if len(item) == 2]
        for edge in all_edges:
            modes += [self[edge[0]][edge[1]]['type']]
        return list(dict.fromkeys(modes))

    def remove_shipment(self, commercial_link):
        """Look for the shipment corresponding to the commercial link
        in any edges and nodes of the main and alternative route,
        and remove it
        """
        route_to_take = commercial_link.route + commercial_link.alternative_route
        for route_segment in route_to_take:
            if len(route_segment) == 2:  # segment is an edge
                if commercial_link.pid in self[route_segment[0]][route_segment[1]]['shipments'].keys():
                    del self[route_segment[0]][route_segment[1]]['shipments'][commercial_link.pid]
            elif len(route_segment) == 1:  # segment is a node
                if commercial_link.pid in self._node[route_segment[0]]['shipments'].keys():
                    del self._node[route_segment[0]]['shipments'][commercial_link.pid]

    def compute_flow_per_segment(self, time_step) -> list:
        """
        Calculate flows of each category and product for each transport edges

        We calculate total flows:
        - for each combination flow_category*product_type
        - for each flow_category
        - for each product_type
        - total of all

        Parameters
        ----------

        Returns
        -------
        flows_per_edge
        """
        flows_per_edge = []
        flows_total = {}
        for edge in self.edges():
            new_data = {
                "time_step": time_step,
                'id': self[edge[0]][edge[1]]['id'],
                'flow_total': 0,
                'flow_total_tons': 0
            }
            # For each shipment, add quantities to relevant categories
            for shipment in self[edge[0]][edge[1]]["shipments"].values():
                flow_name = 'flow_' + shipment['flow_category'] + '_' + shipment['product_type']
                add_or_append_to_dict(new_data, flow_name, shipment['quantity'])
                add_or_append_to_dict(new_data, "flow_" + shipment['flow_category'], shipment['quantity'])
                add_or_append_to_dict(new_data, "flow_" + shipment['product_type'], shipment['quantity'])
                new_data['flow_total'] += shipment['quantity']
                new_data['flow_total_tons'] += shipment['tons']
                add_or_append_to_dict(flows_total, shipment['flow_category'], shipment['quantity'])
                add_or_append_to_dict(flows_total, shipment['flow_category'] + "*km",
                                      shipment['quantity'] * self[edge[0]][edge[1]]["km"])
                add_or_append_to_dict(flows_total, shipment['flow_category'] + "_tons",
                                      shipment['tons'])
                add_or_append_to_dict(flows_total, shipment['flow_category'] + "_tons*km",
                                      shipment['tons'] * self[edge[0]][edge[1]]["km"])
            flows_per_edge += [new_data]
        logging.info(flows_total)
        return flows_per_edge

    def reinitialize_flows_and_disruptions(self):
        for node in self.nodes:
            self.nodes[node]['disruption_duration'] = 0
            self.nodes[node]['shipments'] = {}
        for edge in self.edges:
            self[edge[0]][edge[1]]['disruption_duration'] = 0
            self[edge[0]][edge[1]]['shipments'] = {}
            # self[edge[0]][edge[1]]['congestion'] = 0
            self[edge[0]][edge[1]]['current_load'] = 0
            self[edge[0]][edge[1]]['overused'] = False



from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from src.network.transport_network import TransportNetwork


class Route(list):
    def __init__(self, node_list: list, transport_network: "TransportNetwork"):
        node_edge_tuple = [[(node_list[0],)]] + \
                          [[(node_list[i], node_list[i + 1]), (node_list[i + 1],)]
                           for i in range(0, len(node_list) - 1)]
        transport_nodes_and_edges = [item for item_tuple in node_edge_tuple for item in item_tuple]
        super().__init__(transport_nodes_and_edges)
        self.transport_nodes_and_edges = transport_nodes_and_edges
        self.transport_nodes = node_list
        self.transport_edges = [item for item in transport_nodes_and_edges if len(item) == 2]
        self.transport_edge_ids = [transport_network[source][target]['id'] for source, target in self.transport_edges]
        self.transport_modes = list(set([transport_network[source][target]['type']
                                         for source, target in self.transport_edges]))
        # self.transport_modes = list(transport_edges.loc[self.transport_edge_ids, 'type'].unique())
        self.cost_per_ton = sum([transport_network[source][target]['cost_per_ton']
                                 for source, target in self.transport_edges])
        # self.cost_per_ton = transport_edges.loc[self.transport_edge_ids, 'cost_per_ton'].sum()
        self.length = sum([transport_network[source][target]['km']
                           for source, target in self.transport_edges])

    def check_edge_in_route(self, route, searched_edge):
        for edge in self.transport_edges:
            if (searched_edge[0] == edge[0]) and (searched_edge[1] == edge[1]):
                return True
        return False

    def sum_indicator(self, transport_network: "TransportNetwork", indicator: str, per_type: bool = False):
        if per_type:
            details = []
            for edge in self.transport_edges:
                new_edge = {'id': transport_network[edge[0]][edge[1]]['id'],
                            'type': transport_network[edge[0]][edge[1]]['type'],
                            'multimodes': transport_network[edge[0]][edge[1]]['multimodes'],
                            'special': transport_network[edge[0]][edge[1]]['special'],
                            indicator: transport_network[edge[0]][edge[1]][indicator]}
                details += [new_edge]
            details = pd.DataFrame(details).fillna('N/A')
            # print(details)
            return details.groupby(['type', 'multimodes', 'special'])[indicator].sum()

        else:
            total_indicator = 0
            for edge in self.transport_edges:
                total_indicator += transport_network[edge[0]][edge[1]][indicator]
            return total_indicator



import logging
from pathlib import Path

import pandas
import pandas as pd
import geopandas as gpd
import numpy as np

from src.agents.household import Household, Households
from src.model.builder_functions import get_index_closest_point, get_long_lat, \
    get_closest_road_nodes
from src.model.basic_functions import rescale_monetary_values
from src.network.mrio import Mrio


def create_households(
        household_table: pd.DataFrame,
        household_sector_consumption: dict
):
    """Create the households

    It uses household_table & household_sector_consumption from defineHouseholds

    Parameters
    ----------
    household_table: pandas.DataFrame
        household_table
    household_sector_consumption: dic
        {<household_id>: {<sector>: <amount>}}

    Returns
    -------
    list of Household
    """

    logging.debug('Creating households')
    household_table = household_table.set_index('id')
    households = Households([
        Household('hh_' + str(i),
                  name=household_table.loc[i, "name"],
                  od_point=household_table.loc[i, "od_point"],
                  long=float(household_table.loc[i, 'long']),
                  lat=float(household_table.loc[i, 'lat']),
                  population=household_table.loc[i, "population"],
                  sector_consumption=household_sector_consumption[i]
                  )
        for i in household_table.index.tolist()
    ])
    logging.info('Households generated')

    return households


def define_households_from_mrio_data(
        sector_table: pd.DataFrame,
        filepath_region_table: Path,
        filtered_sectors: list,
        local_demand_cutoff: float,
        transport_nodes: gpd.GeoDataFrame,
        time_resolution: str,
        target_units: str,
        input_units: str
):
    # Adrien's function for Global model
    # household_table = gpd.read_file(filepath_region_table)
    # household_table = location_table.copy(deep=False)
    # TODO ajouter la demande finale par secteur country_sector_table
    final_demand = sector_table["final_demand"]
    # household_table = household_table.stack() #.transpose()

    # TEST

    household_table = gpd.read_file(filepath_region_table)

    # Add final demand
    final_demand = sector_table.loc[sector_table['sector'].isin(filtered_sectors), ['sector', 'final_demand']]
    final_demand_as_row = final_demand.set_index('sector').transpose()

    # Duplicate rows
    household_table = pd.concat([household_table] * len(final_demand_as_row))

    # Align index and concat
    household_table.index = final_demand_as_row.index
    household_table = pd.concat([household_table, final_demand_as_row], axis=1)

    # Reshape the household table
    household_table = household_table.stack().reset_index()
    household_table.columns = ['region', 'column', 'value']

    ## Fin TEST

    # C. Create one household per OD point
    logging.info('Assigning households to od-points')
    dic_select_region_to_points = household_table.set_index('region')['geometry'].to_dict()
    # Select road node points
    road_nodes = transport_nodes[transport_nodes['type'] == "roads"]
    # Create dic
    dic_region_to_road_node_id = {
        region: road_nodes.loc[get_index_closest_point(point, road_nodes), 'id']
        for region, point in dic_select_region_to_points.items()
    }
    # Map household to closest road nodes
    household_table['od_point'] = household_table['region'].map(dic_region_to_road_node_id)

    # Combine households that are in the same od-point
    household_table = household_table \
        .drop(columns=['geometry']) \
        .groupby(['region', 'od_point'], as_index=False) \
        .sum()
    logging.info(str(household_table.shape[0]) + ' od-point selected for demand')

    # D. Filter out small demand
    if local_demand_cutoff > 0:
        household_table[filtered_sectors] = household_table[filtered_sectors].mask(
            household_table[filtered_sectors] < local_demand_cutoff
        )
    # info
    logging.info('Create ' + str(household_table.shape[0]) + " households in " +
                 str(household_table['od_point'].nunique()) + ' od points')
    for sector in filtered_sectors:
        logging.info('Sector ' + sector + ": create " +
                     str((~household_table[sector].isnull()).sum()) +
                     " buying households that covers " +
                     "{:.0f}%".format(
                         household_table[sector].sum() \
                         / sector_table.set_index('sector').loc[sector, 'final_demand'] * 100
                     ) + " of total final demand"
                     )
    if (household_table[filtered_sectors].sum(axis=1) == 0).any():
        logging.warning('Some households have no purchase plan because of all their sectoral demand below cutoff!')

    # E. Add information required by the createHouseholds function
    # add long lat
    od_point_table = road_nodes[road_nodes['id'].isin(household_table['od_point'])].copy()
    od_point_table['long'] = od_point_table.geometry.x
    od_point_table['lat'] = od_point_table.geometry.y
    road_node_id_to_longlat = od_point_table.set_index('id')[['long', 'lat']]
    household_table['long'] = household_table['od_point'].map(road_node_id_to_longlat['long'])
    household_table['lat'] = household_table['od_point'].map(road_node_id_to_longlat['lat'])
    # add id
    household_table['id'] = list(range(household_table.shape[0]))

    # F. Create purchase plan per household
    # rescale according to time resolution
    household_table[filtered_sectors] = rescale_monetary_values(
        household_table[filtered_sectors],
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    # to dict
    household_sector_consumption = household_table.set_index('id')[filtered_sectors].to_dict(orient='index')
    # remove nan values
    household_sector_consumption = {
        i: {
            sector: amount
            for sector, amount in purchase_plan.items()
            if ~np.isnan(amount)
        }
        for i, purchase_plan in household_sector_consumption.items()
    }

    return household_table, household_sector_consumption


def define_households_from_mrio(
        filepath_mrio: Path,
        filepath_region_table: Path,
        transport_nodes: gpd.GeoDataFrame,
        time_resolution: str,
        target_units: str,
        input_units: str
):
    # Load mrio
    mrio = Mrio.load_mrio_from_filepath(filepath_mrio)
    # mrio = pd.read_csv(filepath_mrio, header=[0, 1], index_col=[0, 1])
    # # Extract region_households
    # region_households = mrio.columns.to_list()
    # region_households = [tup for tup in region_households if tup[1] == 'final_demand']

    # Create household table
    household_table = pd.DataFrame({
        'tuple': mrio.region_households,
        'region': [tup[0] for tup in mrio.region_households],
        'name': [tup[0] + '_households' for tup in mrio.region_households]
    })
    # household_table = pd.DataFrame({"name": region_households})
    # household_table['region'] = household_table['name'].str.extract('([0-9]*)-H')
    # household_table['region'] = household_table['name'].str.extract('([A-Z]{3})_H')
    logging.info(f"Select {household_table.shape[0]} households in {household_table['region'].nunique()} regions")

    # Identify OD point
    household_table['od_point'] = get_closest_road_nodes(household_table['region'], transport_nodes,
                                                         filepath_region_table)

    # Add long lat
    long_lat = get_long_lat(household_table['od_point'], transport_nodes)
    household_table['long'] = long_lat['long']
    household_table['lat'] = long_lat['lat']

    # Add id
    household_table['id'] = range(household_table.shape[0])

    # Add population
    region_table = gpd.read_file(filepath_region_table)
    household_table['population'] = household_table['region'].map(region_table.set_index('region')['population'])

    # Identify final demand per region_sector
    consumption = mrio[mrio.region_households].copy(deep=True)   # TODO to put in the class Mrio
    consumption.index = ['_'.join(tup) for tup in consumption.index]
    consumption.columns = [tup[0] + '_households' for tup in mrio.region_households]
    consumption.columns = consumption.columns.map(household_table.set_index('name')['id'])
    consumption = rescale_monetary_values(
        consumption,
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    # col_final_demand = [col for col in mrio.columns if col[-2:] == "-H"]  #TODO too specific
    # col_final_demand = [col for col in mrio.columns if col[-2:] == "_H"]  #TODO too specific
    # household_sector_consumption = mrio[col_final_demand].stack().reset_index() \
    #     .groupby('level_1') \
    #     .apply(lambda df: df.set_index('level_0')[0].to_dict()) \
    #     .to_dict()
    # Replace name by id :TODO use name as id to makes this unecessary
    # dic_name_to_id = household_table.set_index('name')['id']
    household_sector_consumption = consumption.to_dict()
    household_sector_consumption = {
        household: {
            key: value
            for key, value in sublist.items()
            if value > 0
        }
        for household, sublist in household_sector_consumption.items()
    }

    # Info
    logging.info(f"Create {household_table.shape[0]} households in {household_table['od_point'].nunique()} od points")

    return household_table, household_sector_consumption


def define_households(
        sector_table: pd.DataFrame,
        filepath_region_data: Path,
        filtered_sectors: list,
        pop_cutoff: float,
        pop_density_cutoff: float,
        local_demand_cutoff: float,
        transport_nodes: gpd.GeoDataFrame,
        time_resolution: str,
        target_units: str,
        input_units: str
):
    '''Define the number of households to model and their purchase plan based on input demographic data
    and filtering options

    Parameters
    ---------
    input_units
    target_units
    time_resolution
    transport_nodes
    local_demand_cutoff
    pop_density_cutoff
    pop_cutoff
    filtered_sectors
    filepath_region_data
    sector_table

    Returns
    -------
    household_table
    household_purchase_plan

    '''
    # A. Filter regions based on density
    # load file
    region_data = gpd.read_file(filepath_region_data)
    # filter & keep household were firms are
    cond = region_data['population'] > 0
    if pop_density_cutoff > 0:
        cond = cond & (region_data['pop_density'] >= pop_density_cutoff)
    if pop_cutoff > 0:
        cond = cond & (region_data['population'] >= pop_cutoff)
    # create household_table
    household_table = region_data.loc[cond, ['population', 'geometry', 'region']].copy()
    logging.info(f"{cond.sum()} regions selected over {region_data.shape[0]} representing "
                 f"{(household_table['population'].sum() / region_data['population'].sum() * 100):.0f}% of population")

    # B. Add final demand
    # Add imports... weird filtering: add if larger than the lowest sectoral final_demand filtered
    min_final_demand_filtered = sector_table.set_index('sector').loc[filtered_sectors, 'final_demand'].min()
    if sector_table.set_index('sector').loc['IMP', 'final_demand'] > min_final_demand_filtered:
        sectors_to_buy_from = filtered_sectors + ['IMP']
    else:
        sectors_to_buy_from = filtered_sectors
    # get final demand for the selected sector
    final_demand = sector_table.loc[sector_table['sector'].isin(sectors_to_buy_from), ['sector', 'final_demand']]

    # put as single row
    final_demand_as_row = final_demand.set_index('sector').transpose()
    # duplicates rows
    final_demand_each_household = pd.concat([final_demand_as_row for i in range(household_table.shape[0])])
    # align index and concat
    final_demand_each_household.index = household_table.index
    # compute final demand per region
    rel_pop = household_table['population'] / region_data['population'].sum()
    final_demand_each_household = final_demand_each_household.multiply(rel_pop, axis='index')
    # add to household table
    household_table = pd.concat([household_table, final_demand_each_household], axis=1)

    # C. Create one household per OD point
    logging.info('Assigning households to od-points')
    dic_select_region_to_points = household_table.set_index('region')['geometry'].to_dict()
    # Select road node points
    road_nodes = transport_nodes[transport_nodes['type'] == "roads"]
    # Create dic
    dic_region_to_road_node_id = {
        region: road_nodes.loc[get_index_closest_point(point, road_nodes), 'id']
        for region, point in dic_select_region_to_points.items()
    }
    # Map household to the closest road nodes
    household_table['od_point'] = household_table['region'].map(dic_region_to_road_node_id)
    # Combine households that are in the same od-point
    household_table = household_table \
        .drop(columns=['geometry']) \
        .groupby(['od_point', 'region'], as_index=False) \
        .sum()
    logging.info(str(household_table.shape[0]) + ' od-point selected for demand')

    # D. Filter out small demand
    if local_demand_cutoff > 0:
        household_table[sectors_to_buy_from] = household_table[sectors_to_buy_from].mask(
            household_table[sectors_to_buy_from] < local_demand_cutoff
        )
    # info
    logging.info(f"Create {household_table.shape[0]} households in {household_table['od_point'].nunique()} od points")
    for sector in sectors_to_buy_from:
        logging.info(f"Sector {sector}: create {(~household_table[sector].isnull()).sum()} households that covers " +
                     "{:.0f}%".format(
                         household_table[sector].sum() \
                         / sector_table.set_index('sector').loc[sector, 'final_demand'] * 100
                     ) + " of total final demand"
                     )
    if (household_table[sectors_to_buy_from].sum(axis=1) == 0).any():
        logging.warning('Some households have no purchase plan because of all their sectoral demand below cutoff!')

    # E. Add information required by the createHouseholds function
    # add long lat
    od_point_table = road_nodes[road_nodes['id'].isin(household_table['od_point'])].copy()
    od_point_table['long'] = od_point_table.geometry.x
    od_point_table['lat'] = od_point_table.geometry.y
    road_node_id_to_long_lat = od_point_table.set_index('id')[['long', 'lat']]
    household_table['long'] = household_table['od_point'].map(road_node_id_to_long_lat['long'])
    household_table['lat'] = household_table['od_point'].map(road_node_id_to_long_lat['lat'])
    # add id
    household_table['id'] = list(range(household_table.shape[0]))
    # # add name, not really useful
    household_table['name'] = "hh_" + household_table['od_point'].astype(str)

    # F. Create purchase plan per household
    # rescale according to time resolution
    household_table[sectors_to_buy_from] = rescale_monetary_values(
        household_table[sectors_to_buy_from],
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    # to dict
    household_sector_consumption = household_table.set_index('id')[sectors_to_buy_from].to_dict(orient='index')
    # remove nan values
    household_sector_consumption = {
        i: {
            sector: amount
            for sector, amount in purchase_plan.items()
            if ~np.isnan(amount)
        }
        for i, purchase_plan in household_sector_consumption.items()
    }

    return household_table, household_sector_consumption


def add_households_for_firms(
        firm_table: pd.DataFrame,
        household_table: pd.DataFrame,
        filepath_region_data: str,
        sector_table: pd.DataFrame,
        filtered_sectors: list,
        time_resolution: str,
        target_units: str,
        input_units: str
):
    """
    We suppose that all firms face the same final demand. i.e., all firms are both B2B and B2C
    We suppose that households only buy more firms located on the same point
    Some firms are located in areas where there are no households
    Therefore, they cannot sell their final demand
    So this function add households in those points
    """

    # A. Examine where we need new households
    cond_no_household = ~firm_table['od_point'].isin(household_table['od_point'])
    logging.info(cond_no_household.sum(), 'firms without local households')
    logging.info(firm_table.loc[cond_no_household, 'od_point'].nunique(), 'od points with firms without households')

    # B. Create new household table
    added_household_table = firm_table[cond_no_household].groupby('od_point', as_index=False)['population'].max()
    od_point_long_lat = firm_table.loc[cond_no_household, ["od_point", 'long', "lat"]].drop_duplicates()
    added_household_table = added_household_table.merge(od_point_long_lat, how='left', on='od_point')

    # B1. Load region data to get tot population
    region_data = gpd.read_file(filepath_region_data)
    tot_pop = region_data['population'].sum()

    # B2. Load sector table to get final demand
    final_demand = sector_table.loc[sector_table['sector'].isin(filtered_sectors), ['sector', 'final_demand']]

    # B3. Add final demand per sector per new household
    # put as single row
    final_demand_as_row = final_demand.set_index('sector').transpose()
    # duplicates rows
    final_demand_each_household = pd.concat([final_demand_as_row for i in range(added_household_table.shape[0])])
    # align index and concat
    final_demand_each_household.index = added_household_table.index
    # compute final demand per commune
    rel_pop = added_household_table['population'] / tot_pop
    final_demand_each_household = final_demand_each_household.multiply(rel_pop, axis='index')
    # keep demand only for firm that are there
    cond_to_reject = (
        firm_table[cond_no_household].groupby(['od_point', 'sector'])['population'].sum().isnull()).unstack('sector')
    cond_to_reject = cond_to_reject.fillna(True)
    cond_to_reject.index = final_demand_each_household.index
    final_demand_each_household = final_demand_each_household.mask(cond_to_reject)
    # add to household table
    added_household_table = pd.concat([added_household_table, final_demand_each_household], axis=1)
    # rescale according to time resolution
    added_household_table[filtered_sectors] = rescale_monetary_values(
        added_household_table[filtered_sectors],
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    # C. Merge
    added_household_table['id'] = [household_table['id'].max() + 1 + i for i in range(added_household_table.shape[0])]
    added_household_table.index = added_household_table['id']
    household_table = pd.concat([household_table, added_household_table], sort=True)
    # household_table.to_csv('tt.csv')

    # D. Log info
    logging.info('Create ' + str(household_table.shape[0]) + " households in " +
                 str(household_table['od_point'].nunique()) + ' od points')
    for sector in filtered_sectors:
        tot_demand = rescale_monetary_values(sector_table.set_index('sector').loc[sector, 'final_demand'],
                                             time_resolution=time_resolution, target_units=target_units,
                                             input_units=input_units)
        logging.info('Sector ' + sector + ": create " +
                     str((~household_table[sector].isnull()).sum()) +
                     " buying households that covers " +
                     "{:.0f}%".format(
                         household_table[sector].sum() \
                         / tot_demand * 100
                     ) + " of total final demand"
                     )
    if (household_table[filtered_sectors].sum(axis=1) == 0).any():
        logging.warning('Some households have no purchase plan!')

    # E. Create household_sector_consumption dic
    # create dic
    household_sector_consumption = household_table.set_index('id')[filtered_sectors].to_dict(orient='index')
    # remove nan values
    household_sector_consumption = {
        i: {
            sector: amount
            for sector, amount in purchase_plan.items()
            if ~np.isnan(amount)
        }
        for i, purchase_plan in household_sector_consumption.items()
    }

    return household_table, household_sector_consumption
